#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""
RemoteServiceAdmin constants and utility functions

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from argparse import ArgumentError
import logging
import threading
import time
import uuid

from pelix import constants

# pylint: disable=W0125
if False:
    # pylint: disable=W0611
    # Ugly trick to enable PyCharm/MyPy typing without making cycle imports
    from typing import Tuple, List, Dict, Any, Optional

    from pelix.framework import Bundle, BundleContext, ServiceReference
    from pelix.rsa.endpointdescription import EndpointDescription

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Standard logging
_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------

# RSA constants, declared in org.osgi.service.remoteserviceadmin.RemoteConstants
ENDPOINT_ID = "endpoint.id"
ENDPOINT_SERVICE_ID = "endpoint.service.id"
ENDPOINT_FRAMEWORK_UUID = "endpoint.framework.uuid"
ENDPOINT_PACKAGE_VERSION_ = "endpoint.package.version."
SERVICE_EXPORTED_INTERFACES = "service.exported.interfaces"
REMOTE_CONFIGS_SUPPORTED = "remote.configs.supported"
REMOTE_INTENTS_SUPPORTED = "remote.intents.supported"
SERVICE_EXPORTED_CONFIGS = "service.exported.configs"
SERVICE_EXPORTED_INTENTS = "service.exported.intents"
SERVICE_EXPORTED_INTENTS_EXTRA = "service.exported.intents.extra"
SERVICE_IMPORTED = "service.imported"
SERVICE_IMPORTED_CONFIGS = "service.imported.configs"
SERVICE_INTENTS = "service.intents"
SERVICE_ID = "service.id"
OBJECT_CLASS = "objectClass"
SERVICE_BUNDLE_ID = "service.bundleid"
INSTANCE_NAME = "instance.name"
SERVICE_RANKING = "service.ranking"
SERVICE_COMPONENT_NAME = "component.name"
SERVICE_COMPONENT_ID = "component.id"
# R7 standardized intents
OSGI_BASIC_INTENT = "osgi.basic"
OSGI_BASIC_TIMEOUT_INTENT = "{0}.{1}".format(OSGI_BASIC_INTENT, "timeout")
OSGI_ASYNC_INTENT = "osgi.async"
OSGI_CONFIDENTIAL_INTENT = "osgi.confidential"
OSGI_PRIVATE_INTENT = "osgi.private"
# List of them
RSA_PROP_NAMES = [
    ENDPOINT_ID,
    ENDPOINT_SERVICE_ID,
    ENDPOINT_FRAMEWORK_UUID,
    SERVICE_EXPORTED_INTERFACES,
    REMOTE_CONFIGS_SUPPORTED,
    REMOTE_INTENTS_SUPPORTED,
    SERVICE_EXPORTED_CONFIGS,
    SERVICE_EXPORTED_INTENTS,
    SERVICE_EXPORTED_INTENTS_EXTRA,
    SERVICE_IMPORTED,
    SERVICE_IMPORTED_CONFIGS,
    SERVICE_INTENTS,
    SERVICE_ID,
    OBJECT_CLASS,
    INSTANCE_NAME,
    SERVICE_RANKING,
    SERVICE_COMPONENT_ID,
    SERVICE_COMPONENT_NAME,
]

# ECF constants
ECF_ENDPOINT_CONTAINERID_NAMESPACE = "ecf.endpoint.id.ns"
ECF_ENDPOINT_ID = "ecf.endpoint.id"
ECF_RSVC_ID = "ecf.rsvc.id"
ECF_ENDPOINT_TIMESTAMP = "ecf.endpoint.ts"
ECF_ENDPOINT_CONNECTTARGET_ID = "ecf.endpoint.connecttarget.id"
ECF_ENDPOINT_IDFILTER_IDS = "ecf.endpoint.idfilter.ids"
ECF_ENDPOINT_REMOTESERVICE_FILTER = "ecf.endpoint.rsfilter"
ECF_SERVICE_EXPORTED_CONTAINER_FACTORY_ARGS = (
    "ecf.exported.containerfactoryargs"
)
ECF_SERVICE_EXPORTED_CONTAINER_CONNECT_CONTEXT = (
    "ecf.exported.containerconnectcontext"
)
ECF_SERVICE_EXPORTED_CONTAINER_ID = "ecf.exported.containerid"
ECF_SERVICE_EXPORTED_ASYNC_INTERFACES = "ecf.exported.async.interfaces"
ECF_SERVICE_EXPORTED_ASYNC_NOPROXY = "ecf.rsvc.async.noproxy"
ECF_SERVICE_ASYNC_RSPROXY_CLASS_ = "ecf.rsvc.async.proxy_"
ECF_ASYNC_INTERFACE_SUFFIX = "Async"
ECF_SERVICE_IMPORTED_VALUETYPE = "ecf.service.imported.valuetype"
ECF_SERVICE_IMPORTED_ENDPOINT_ID = ENDPOINT_ID
ECF_SERVICE_IMPORTED_ENDPOINT_SERVICE_ID = ENDPOINT_SERVICE_ID
ECF_OSGI_ENDPOINT_MODIFIED = "ecf.osgi.endpoint.modified"
ECF_OSGI_CONTAINER_ID_NS = "ecf.osgi.ns"
# List
ECFPROPNAMES = [
    ECF_ENDPOINT_CONTAINERID_NAMESPACE,
    ECF_ENDPOINT_ID,
    ECF_RSVC_ID,
    ECF_ENDPOINT_TIMESTAMP,
    ECF_ENDPOINT_CONNECTTARGET_ID,
    ECF_ENDPOINT_IDFILTER_IDS,
    ECF_ENDPOINT_REMOTESERVICE_FILTER,
    ECF_SERVICE_EXPORTED_ASYNC_INTERFACES,
    ECF_SERVICE_IMPORTED_VALUETYPE,
]

ERROR_EP_ID = "0"
ERROR_NAMESPACE = "org.eclipse.ecf.core.identity.StringID"
ERROR_IMPORTED_CONFIGS = ["import.error.config"]
ERROR_ECF_EP_ID = "export.error.id"
DEFAULT_EXPORTED_CONFIGS = ["ecf.xmlrpc.server"]

# ------------------------------------------------------------------------------
## RSA Service API
# ------------------------------------------------------------------------------

SERVICE_REMOTE_SERVICE_ADMIN = "pelix.rsa.remoteserviceadmin"


class RemoteServiceAdmin(object):
    """
    RSA service specification.  This specification is the core service
    implemented by the RSA package.  See the RemoteServiceAdmin
    class below for method documentation.
    """

    def get_exported_services(self):
        # type: () -> List[ExportReference]
        """
        Get services previously exported by this RSA implementation.  Will
        not return None, but may return empty list.

        :return list of ExportReference instances.  See ExportReference class.
        """
        raise Exception(
            "{0}.get_exported_services not implemented".format(self)
        )

    def get_imported_endpoints(self):
        # type: () -> List[ImportReference]
        """
        Get services previously imported by this RSA implementation.  Will
        not return None, but may return empty list.

        :return list of ImportReference instances.  See ImpportReference class.
        """
        raise Exception(
            "{0}.get_imported_endpoints not implemented".format(self)
        )

    def export_service(self, service_ref, overriding_props=None):
        # type: (ServiceReference, Dict[str, Any]) -> List[ExportRegistration]
        # pylint: disable=W0613
        """
        Export a given service_ref (ServiceReference) using overriding_props
        dictionary. service_ref must not be None and must be of type
        ServiceReference.  If overriding_props is provided,
        then the properties set in overriding_props will override those
        from service_ref.get_properties().  The RSA-specified props
        (e.g. service.exported.interfaces,service.exported.configs,
        service.exported.intents) will then be used to determine whether
        to export the service, and which ExportDistributionProvider
        services to use for exporting the service.

        :param service_ref a service to export.  Must not be None and
        must be of type ServiceReference
        :param overriding_props if not None, any props given will
        override those provided in service_ref
        :return list of ExportRegistration instances.  See ExportRegistration
        class.
        """
        raise Exception("{0}.export_service not implemented".format(self))

    def import_service(self, endpoint_description):
        # type: (EndpointDescription) -> ImportRegistration
        # pylint: disable=W0613
        """
        Import a given endpoint_description.  Must not be None, and must
        be of type EndpointDescription.  The endpoint_description props
        (e.g. service.exported.configs, service.imported.configs, etc,
        service.intents) will then be used to determine whether
        to import the service, and which ImportDistributionProvider
        service to use for importing the service.

        :param endpoint_description to import.  Must not be None.
        Must be of type EndpointDescription
        :return a single ImportRegistration used to import the
        endpoint_description.  See ImportRegistration class
        """
        raise Exception("{0}.import_service not implemented".format(self))


# ------------------------------------------------------------------------------


class ExportRegistration(object):
    """
    Declaration of ExportRegistration signature.  Instance of this class
    are returned from RemoteServiceAdmin.export_service to describe the
    exported service.
    """

    def get_export_reference(self):
        # type: () -> ExportReference
        """
        Get the ExportReference associated with this ExportRegistration.  Will
        be None if this registration has been previously closed.  See
        ExportReference class.

        :return ExportReference associated with this registration, or None
        """
        raise Exception("{0}.get_export_reference not implemented".format(self))

    def get_export_container_id(self):
        # type: () -> Tuple[str, str]
        """
        Get the exporting container id of form
        tuple(namespace(string), containerid(string)).
        For example:  ('ecf.namespace.xmlrpc','http://localhost/xml-rpc').
        Will not return None.

        :return: exporting container id of form:
                 tuple(namespace(string),containerid(string))
        """
        raise Exception(
            "{0}.get_export_container_id not implemented".format(self)
        )

    def get_remoteservice_id(self):
        # type: () -> Tuple[Tuple[str, str], int]
        """
        Get the exporting remoteservice id of form:
        tuple(containerid,rsid(int)),
        with containerid of form returned from get_export_container_id.

        For example:
        (('ecf.namespace.xmlrpc','http://localhost/xml-rpc'),1).

        Will not be None.

        :return: exporting remote service id of form:
                 tuple(containerid,rsid(int))
        """
        raise Exception("{0}.get_remoteservice_id not implemented".format(self))

    def get_reference(self):
        # type: () -> ServiceReference
        """
        Get the ServiceReference associated with this ExportRegistration.  Will
        be None if the ExportRegistration has been closed, or if an exception
        occurred on attempted export

        :return ServiceReference associated with this ExportRegistration or
        None if this has been previously closed, or if an exception occurred
        on attempted export.
        """
        raise Exception("{0}.get_reference not implemented".format(self))

    def get_exception(self):
        # type: () -> Optional[Tuple[Any, Any, Any]]
        """
        Get any exception associated with the attempted export.  If not None,
        will be of form:  tuple(exc_type,exc_msg,exc_stack).  For example:
        (SelectContainerException,'No container available',stack_trace).

        :return exception tuple of form: tuple(exc_type,exc_msg,exc_stack) or
        None if no exception occurred during the export associated with this
        ExportRegistration.
        """
        raise Exception("{0}.get_exception not implemented".format(self))

    def get_description(self):
        # type: () -> EndpointDescription
        """
        Get EndpointDescription associated with this ExportRegistration.
        Will not be None.  See EndpointDescription class.

        :return EndpointDescription associated with this registration
        """
        raise Exception("{0}.get_description not implemented".format(self))

    def match_sr(self, svc_ref, cid=None):
        # type: (ServiceReference, Optional[Tuple[str, str]] ) -> bool
        # pylint: disable=W0613
        """
        Checks if this export registration matches the given service reference

        :param svc_ref: A service reference
        :param cid: A container ID
        :return: True if the service matches this export registration
        """
        raise Exception("{0}.get_description not implemented".format(self))

    def update(self, properties):
        # type: (Dict[str, Any]) -> EndpointDescription
        # pylint: disable=W0613
        """
        Updates ExportRegistration with new properties.

        :param properties a dictionary of new properties.  May be None.
        :return: EndpointDescription for ExportRegistration, or None if not
        updated.
        """
        raise Exception("{0}.update not implemented".format(self))

    def close(self):
        # type: () -> None
        """
        Close this ExportRegistration.  If called after having been previously
        called, will have no effect.
        """
        raise Exception("{0}.close not implemented".format(self))


# ------------------------------------------------------------------------------


class ExportReference(object):
    """
    Declaration of ExportReference signature.  Instance of this class
    are returned from ExportRegistration.get_export_reference().
    See ExportRegistration.get_export_reference or RemoteServiceAdmin.
    get_exported_services.
    """

    def get_export_container_id(self):
        # type: () -> Tuple[str, str]
        """
        Get the exporting container id of form
        tuple(namespace(string),containerid(string)).

        For example:  ('ecf.namespace.xmlrpc','http://localhost/xml-rpc').
        Will be None if this reference has previously been closed.

        :return: exporting container id of form:
                 tuple(namespace(string),containerid(string))
        """
        raise Exception(
            "{0}.get_export_container_id not implemented".format(self)
        )

    def get_remoteservice_id(self):
        # type: () -> Tuple[Tuple[str, str], int]
        """
        Get the exporting remoteservice id of form:
        tuple(containerid,rsid(int)), with containerid of form returned
        from get_export_container_id.

        For example:
        (('ecf.namespace.xmlrpc','http://localhost/xml-rpc'),1).

        Will be None if this reference has previously been closed.

        :return: exporting remote service id of form:
                 tuple(containerid,rsid(int))
        """
        raise Exception("{0}.get_remoteservice_id not implemented".format(self))

    def get_reference(self):
        # type: () -> ServiceReference
        """
        Get the ServiceReference associated with this ExportReference.  Will
        be None if the ExportReference has been closed, or if an exception
        occurred during the attempted export.

        :return: ServiceReference associated with this ExportReference or
        None if this has been previously closed, or if an exception occurred
        during attempted export.
        """
        raise Exception("{0}.get_reference not implemented".format(self))

    def get_description(self):
        # type: () -> EndpointDescription
        """
        Get EndpointDescription associated with this ExportReference.
        Will not be None.  See EndpointDescription class.  Will be None
        if ExportReference has previously been closed.

        :return: EndpointDescription associated with this reference or
        None if reference has previously been closed.
        """
        raise Exception("{0}.get_description not implemented".format(self))

    def get_exception(self):
        # type: () -> Tuple[Any, Any, Any]
        """
        Get any exception associated with the attempted export.  If not None,
        will be of form:  tuple(exc_type,exc_msg,exc_stack).  For example:
        (SelectContainerException,'No container available',stack_trace).

        :return: exception tuple of form: tuple(exc_type,exc_msg,exc_stack) or
        None if no exception occurred during the export associated with this
        ExportReference, or if previously closed.
        """
        raise Exception("{0}.get_exception not implemented".format(self))

    def update(self, properties):
        # type: (Dict[str, Any]) -> EndpointDescription
        # pylint: disable=W0613
        """
        Update the service properties of the exported service.

        :param properties: Dictionary of new properties. Should not be None
        :return EndpointDescription associated with existing or None
        if reference previously closed or exception occurred during
        export.
        """
        raise Exception("{0}.update not implemented".format(self))

    def close(self, export_reg):
        # type: (ExportRegistration) -> None
        # pylint: disable=W0613
        """
        Close this ExportRegistration.  If called after having been previously
        called, will have no effect.
        """
        raise Exception("{0}.close not implemented".format(self))


# ------------------------------------------------------------------------------


class ImportRegistration(object):
    """
    Declaration of ImportRegistration signature.  Instance of this class
    are returned from RemoteServiceAdmin.import_service to allow the
    imported service to be managed.
    """

    def get_import_reference(self):
        # type: () -> ImportReference
        """
        Get the ImportReference associated with this ImportRegistration.  Will
        be None if this registration has been previously closed.  See
        ImportReference class.

        :return ImportReference associated with this registration, or None
        """
        raise Exception("{0}.get_import_reference not implemented".format(self))

    def get_import_container_id(self):
        # type: () -> Tuple[str, str]
        """
        Get the importing container id of form
        tuple(namespace(string),containerid(string)).

        For example:
        ('ecf.namespace.xmlrpc','123e4567-e89b-42d3-a456-556642440000')

        Will not return None.

        :return: importing container id of form:
                 tuple(namespace(string),containerid(string))
        """
        raise Exception(
            "{0}.get_import_container_id not implemented".format(self)
        )

    def get_export_container_id(self):
        # type: () -> Tuple[str, str]
        """
        Get the exporting container id of form
        tuple(namespace(string),containerid(string)).

        For example:
        ('ecf.namespace.xmlrpc','http://localhost/xml-rpc')

        Will not return None.

        :return: importing container id of form:
                 tuple(namespace(string),containerid(string))
        """
        raise Exception(
            "{0}.get_export_container_id not implemented".format(self)
        )

    def get_remoteservice_id(self):
        # type: () -> Tuple[Tuple[str, str], int]
        """
        Get the exporting remoteservice id of form:
        tuple(containerid,rsid(int)), with containerid of form returned from
        get_export_container_id.

        For example: (('ecf.namespace.xmlrpc','http://localhost/xml-rpc'),1).
        Will not be None.

        :return: exporting remote service id of form:
                 tuple(containerid,rsid(int))
        """
        raise Exception("{0}.get_remoteservice_id not implemented".format(self))

    def get_reference(self):
        # type: () -> Optional[ServiceReference]
        """
        Get the ServiceReference associated with this ImportRegistration.
        Will be None if the ImportRegistration has been closed, or if an
        exception occurred on attempted export.

        :return: ServiceReference associated with this ImportRegistration or
        None if this has been previously closed, or if an exception occurred
        on attempted export.
        """
        raise Exception("{0}.get_reference not implemented".format(self))

    def get_exception(self):
        # type: () -> Optional[Tuple[Any, Any, Any]]
        """
        Get any exception associated with the attempted import.  If not None,
        will be of form:  tuple(exc_type,exc_msg,exc_stack).

        For example:
        (SelectContainerException,'No container available',stack_trace)

        :return: exception tuple of form: tuple(exc_type,exc_msg,exc_stack) or
        None if no exception occurred during the import associated with this
        ImportRegistration.
        """
        raise Exception("{0}.get_exception not implemented".format(self))

    def get_description(self):
        # type: () -> Optional[EndpointDescription]
        """
        Get EndpointDescription associated with this ImportRegistration.
        Will not be None.  See EndpointDescription class.

        :return: EndpointDescription associated with this registration
        """
        raise Exception("{0}.get_description not implemented".format(self))

    def update(self, endpoint_description):
        # type: (EndpointDescription) -> boolean
        # pylint: disable=W0613
        """
        Update the service properties of the imported service.

        :param endpoint_description: EndpointDescription for updated endpoint.
                                     Will not be None.
        :return: True if update completed successfully, False if not.
        """
        raise Exception("{0}.update not implemented".format(self))

    def close(self):
        # type: () -> None
        """
        Close this ImportRegistration.  If called after having been previously
        called, will have no effect.
        """
        raise Exception("{0}.close not implemented".format(self))


# ------------------------------------------------------------------------------


class ImportReference:
    """
    Declaration of ImportReference signature.  Instance of this class
    are returned from ImportRegistration.get_export_reference().
    See ImportRegistration.get_export_reference or RemoteServiceAdmin.
    get_imported_endpoints.
    """

    def get_import_container_id(self):
        # type: () -> Tuple[str, str]
        """
        Get the importing container id of form
        tuple(namespace(string),containerid(string)).

        For example:
        ('ecf.namespace.xmlrpc','123e4567-e89b-42d3-a456-556642440000')

        Will be None if this reference has previously been closed.

        :return: exporting container id of form:
                 tuple(namespace(string),containerid(string))
        """
        raise Exception(
            "{0}.get_export_container_id not implemented".format(self)
        )

    def get_export_container_id(self):
        # type: () -> Tuple[str, str]
        """
        Get the exporting container id of form
        tuple(namespace(string),containerid(string)).

        For example:
        ('ecf.namespace.xmlrpc','http://localhost/xml-rpc')

        Will be None if this reference has previously been closed.

        :return: exporting container id of form:
                 tuple(namespace(string),containerid(string))
        """
        raise Exception(
            "{0}.get_export_container_id not implemented".format(self)
        )

    def get_remoteservice_id(self):
        # type: () -> Tuple[Tuple[str, str], int]
        """
        Get the importing remoteservice id of form:
        tuple(containerid,rsid(int)), with containerid of form returned from
        get_import_container_id.

        For example:
        (('ecf.namespace.xmlrpc','http://localhost/xml-rpc'),1).

        Will be None if this reference has previously been closed.

        :return: importing remote service id of form:
                 tuple(containerid,rsid(int))
        """
        raise Exception("{0}.get_remoteservice_id not implemented".format(self))

    def get_reference(self):
        # type: () -> ServiceReference
        """
        Get the ServiceReference of proxy associated with this ImportReference.
        Will be None if the ImportReference has been closed, or if an exception
        occurred during the attempted import.

        :return: ServiceReference associated with this ImportReference or
        None if this has been previously closed, or if an exception occurred
        during attempted export.
        """
        raise Exception("{0}.get_reference not implemented".format(self))

    def get_description(self):
        # type: () -> EndpointDescription
        """
        Get EndpointDescription associated with this ImportReference.
        Will not be None.  See EndpointDescription class.  Will be None
        if ImportReference has previously been closed.

        :return: EndpointDescription associated with this reference or
        None if reference has previously been closed.
        """
        raise Exception("{0}.get_description not implemented".format(self))

    def get_exception(self):
        # type: () -> Optional[Tuple[Any, Any, Any]]
        """
        Get any exception associated with the attempted import.  If not None,
        will be of form:  tuple(exc_type,exc_msg,exc_stack).  For example:
        (SelectContainerException,'No container available',stack_trace).

        :return: exception tuple of form: tuple(exc_type,exc_msg,exc_stack) or
        None if no exception occurred during the import associated with this
        ImportReference, or if previously closed.
        """
        raise Exception("{0}.get_exception not implemented".format(self))

    def update(self, endpoint):
        # type: (EndpointDescription) -> Optional[EndpointDescription]
        # pylint: disable=W0613
        """
        Update the service properties of the imported service.

        :param endpoint: Updated description of the endpoint.
                         Should not be None
        :return: EndpointDescription associated with existing or None if
                 reference previously closed or exception occurred during
                 import.
        """
        raise Exception("{0}.update not implemented".format(self))

    def close(self, import_reg):
        # type: (ImportRegistration) -> None
        # pylint: disable=W0613
        """
        Close this ImportReference.  If called after having been previously
        called, will have no effect.
        """
        raise Exception("{0}.close not implemented".format(self))


# ------------------------------------------------------------------------------
# Remote Service Admin Event Listener service specification.  Instances
# of this service are synchronously called by the RemoteServiceAdmin service
# to notify of RemoteServiceAdminEvents.  See RemoteServiceAdminEvent
# class for types of events and data associated with event.  All services
# registered with this service specification will have their remote_admin_event
# method called by RSA after the appropriate event.
SERVICE_RSA_EVENT_LISTENER = "pelix.rsa.remoteserviceadmineventlistener"


class RemoteServiceAdminListener(object):
    # pylint: disable=R0903
    """
    Remote service admin listener service interface.  Services
    registered with this as service specification will have this method
    called synchronously by the RSA implementation for notification
    of RSA events.  The event parameter will be of type
    RemoteServiceAdminEvent (see below).
    """

    def remote_admin_event(self, rsa_event):
        # type: (RemoteServiceAdminEvent) -> None
        """
        Method called by RSA implementation when RSA events occur.   See
        RemoteServiceAdminEvent above for types of events, and the information
        in each event.

        :param rsa_event the RemoteServiceAdminEvent instance.  Will not
        be None
        """
        pass


# ------------------------------------------------------------------------------
class RemoteServiceAdminEvent(object):
    """
    Remote service admin event instances are delivered to
    RemoteServiceAdminListener service instances when events of the types
    listed below occur... e.g. IMPORT_REGISTRATION when a successful import
    occurs, EXPORT_REGISTRATION when a successful export occurs, etc.
    """

    IMPORT_REGISTRATION = 1
    EXPORT_REGISTRATION = 2
    EXPORT_UNREGISTRATION = 3
    IMPORT_UNREGISTRATION = 4
    IMPORT_ERROR = 5
    EXPORT_ERROR = 6
    EXPORT_WARNING = 7
    IMPORT_WARNING = 8
    IMPORT_UPDATE = 9
    EXPORT_UPDATE = 10

    @classmethod
    def fromimportreg(cls, bundle, import_reg):
        # type: (Bundle, ImportRegistration) -> RemoteServiceAdminEvent
        """
        Creates a RemoteServiceAdminEvent object from an ImportRegistration
        """
        exc = import_reg.get_exception()
        if exc:
            return RemoteServiceAdminEvent(
                RemoteServiceAdminEvent.IMPORT_ERROR,
                bundle,
                import_reg.get_import_container_id(),
                import_reg.get_remoteservice_id(),
                None,
                None,
                exc,
                import_reg.get_description(),
            )

        return RemoteServiceAdminEvent(
            RemoteServiceAdminEvent.IMPORT_REGISTRATION,
            bundle,
            import_reg.get_import_container_id(),
            import_reg.get_remoteservice_id(),
            import_reg.get_import_reference(),
            None,
            None,
            import_reg.get_description(),
        )

    @classmethod
    def fromexportreg(cls, bundle, export_reg):
        # type: (Bundle, ExportRegistration) -> RemoteServiceAdminEvent
        """
        Creates a RemoteServiceAdminEvent object from an ExportRegistration
        """
        exc = export_reg.get_exception()
        if exc:
            return RemoteServiceAdminEvent(
                RemoteServiceAdminEvent.EXPORT_ERROR,
                bundle,
                export_reg.get_export_container_id(),
                export_reg.get_remoteservice_id(),
                None,
                None,
                exc,
                export_reg.get_description(),
            )

        return RemoteServiceAdminEvent(
            RemoteServiceAdminEvent.EXPORT_REGISTRATION,
            bundle,
            export_reg.get_export_container_id(),
            export_reg.get_remoteservice_id(),
            None,
            export_reg.get_export_reference(),
            None,
            export_reg.get_description(),
        )

    @classmethod
    def fromexportupdate(cls, bundle, export_reg):
        # type: (Bundle, ExportRegistration) -> RemoteServiceAdminEvent
        """
        Creates a RemoteServiceAdminEvent object from the update of an
        ExportRegistration
        """
        exc = export_reg.get_exception()
        if exc:
            return RemoteServiceAdminEvent(
                RemoteServiceAdminEvent.EXPORT_ERROR,
                bundle,
                export_reg.get_export_container_id(),
                export_reg.get_remoteservice_id(),
                None,
                export_reg.get_export_reference(),
                None,
                export_reg.get_description(),
            )

        return RemoteServiceAdminEvent(
            RemoteServiceAdminEvent.EXPORT_UPDATE,
            bundle,
            export_reg.get_export_container_id(),
            export_reg.get_remoteservice_id(),
            None,
            export_reg.get_export_reference(),
            None,
            export_reg.get_description(),
        )

    @classmethod
    def fromimportupdate(cls, bundle, import_reg):
        # type: (Bundle, ImportRegistration) -> RemoteServiceAdminEvent
        """
        Creates a RemoteServiceAdminEvent object from the update of an
        ImportRegistration
        """
        exc = import_reg.get_exception()
        if exc:
            return RemoteServiceAdminEvent(
                RemoteServiceAdminEvent.IMPORT_ERROR,
                bundle,
                import_reg.get_import_container_id(),
                import_reg.get_remoteservice_id(),
                None,
                None,
                exc,
                import_reg.get_description(),
            )

        return RemoteServiceAdminEvent(
            RemoteServiceAdminEvent.IMPORT_UPDATE,
            bundle,
            import_reg.get_import_container_id(),
            import_reg.get_remoteservice_id(),
            import_reg.get_import_reference(),
            None,
            None,
            import_reg.get_description(),
        )

    @classmethod
    def fromimportunreg(
        cls, bundle, cid, rsid, import_ref, exception, endpoint
    ):
        # type: (Bundle, Tuple[str, str], Tuple[Tuple[str, str], int], ImportReference, Optional[Tuple[Any, Any, Any]], EndpointDescription) -> RemoteServiceAdminEvent
        """
        Creates a RemoteServiceAdminEvent object from the departure of an
        ImportRegistration
        """
        return RemoteServiceAdminEvent(
            typ=RemoteServiceAdminEvent.IMPORT_UNREGISTRATION,
            bundle=bundle,
            cid=cid,
            rsid=rsid,
            import_ref=import_ref,
            exception=exception,
            endpoint=endpoint,
        )

    @classmethod
    def fromexportunreg(
        cls, bundle, exporterid, rsid, export_ref, exception, endpoint
    ):
        # type: (Bundle, Tuple[str, str], Tuple[Tuple[str, str], int], ExportReference, Optional[Tuple[Any, Any, Any]], EndpointDescription) -> RemoteServiceAdminEvent
        """
        Creates a RemoteServiceAdminEvent object from the departure of an
        ExportRegistration
        """
        return RemoteServiceAdminEvent(
            typ=RemoteServiceAdminEvent.EXPORT_UNREGISTRATION,
            bundle=bundle,
            cid=exporterid,
            rsid=rsid,
            export_ref=export_ref,
            exception=exception,
            endpoint=endpoint,
        )

    @classmethod
    def fromimporterror(cls, bundle, importerid, rsid, exception, endpoint):
        # type: (Bundle, Tuple[str, str], Tuple[Tuple[str, str], int], Optional[Tuple[Any, Any, Any]], EndpointDescription) -> RemoteServiceAdminEvent
        """
        Creates a RemoteServiceAdminEvent object from an import error
        """
        return RemoteServiceAdminEvent(
            RemoteServiceAdminEvent.IMPORT_ERROR,
            bundle,
            importerid,
            rsid,
            None,
            None,
            exception,
            endpoint,
        )

    @classmethod
    def fromexporterror(cls, bundle, exporterid, rsid, exception, endpoint):
        # type: (Bundle, Tuple[str, str], Tuple[Tuple[str, str], int], Optional[Tuple[Any, Any, Any]], EndpointDescription) -> RemoteServiceAdminEvent
        """
        Creates a RemoteServiceAdminEvent object from an export error
        """
        return RemoteServiceAdminEvent(
            RemoteServiceAdminEvent.EXPORT_ERROR,
            bundle,
            exporterid,
            rsid,
            None,
            None,
            exception,
            endpoint,
        )

    def __init__(
        self,
        typ,
        bundle,
        cid,
        rsid,
        import_ref=None,
        export_ref=None,
        exception=None,
        endpoint=None,
    ):
        # type: (int, Bundle, Tuple[str, str], Tuple[Tuple[str, str], int], Optional[ImportReference], Optional[ExportReference], Optional[Tuple[Any, Any, Any]], EndpointDescription) -> None
        self._type = typ
        self._bundle = bundle
        self._cid = cid
        self._rsid = rsid
        self._import_ref = import_ref
        self._export_ref = export_ref
        self._exception = exception
        self._ed = endpoint

    def get_description(self):
        # type: () -> Optional[EndpointDescription]
        """
        Get the EndpointDescription associated with this event.
        Will not be None

        :return EndpointDescription associated with this event
        """
        return self._ed

    def get_container_id(self):
        # type: () -> Tuple[str, str]
        """
        Get the container id of form tuple/2 (namespace,id) where
        both namespace and id are strings. Will not be none.

        :return tuple of namespace,id strings for the Container used
        for export (ExportContainer) or import (ImportContainer).
        """
        return self._cid

    def get_remoteservice_id(self):
        # type: () -> Tuple[Tuple[str, str], int]
        """
        Get the remote service id of form:  tuple(tuple(namespace,id),rsid)
        where rsid is int and (namespace,id) are as returned from
        get_container_id.  This identifies the *exporting* remote
        service id, so the container id will be the same for
        export and different for import events.

        :return tuple(tuple(namespace,id),rsid) to represent the
        remote service id.
        """
        return self._rsid

    def get_type(self):
        # type: () -> int
        """
        Get type of RSA event.  Will be one of the constants
        RemoteServiceAdminEvent.IMPORT_REGISTRATION,EXPORT_REGISTRATION, etc.

        :return rsa event type (int)
        """
        return self._type

    def get_source(self):
        # type: () -> Bundle
        """
        Get the Bundle source for this event.  Will usually be
        the pelix.rsa.remoteserviceadmin event.  Will not be
        None.

        :return source bundle for this event.
        """
        return self._bundle

    def get_import_ref(self):
        # type: () -> Optional[ImportReference]
        """
        Get ImportReference instance associated with this event.
        Will be None if type is IMPORT_*.

        :return import reference associated with this event
        """
        return self._import_ref

    def get_export_ref(self):
        # type: () -> Optional[ExportReference]
        """
        Get ExportReference instance associated with this event.
        Will be None if type is EXPORT_*.

        :return export reference associated with this event
        """
        return self._export_ref

    def get_exception(self):
        # type: () -> Optional[Tuple[Any, Any, Any]]
        """
        Get exception in tuple(exc_type,exc_name,traceback) form.
        If None, no exception occurred in RSA import/export. If
        not None, then an exception occurred and the EVENT_TYPE will
        be *ERROR
        """
        return self._exception


def create_uuid():
    # type: () -> str
    """
    Generates a UUID 4

    :return: A string UUID
    """
    return str(uuid.uuid4())


def create_uuid_uri():
    # type: () -> str
    """
    Generates a UUID URI

    :return: A uuid:<uuid> string
    """
    return "uuid:" + create_uuid()


def time_since_epoch():
    """
    Gives a timestamp floored to last second
    """
    # type: () -> int
    return int(time.time() - 1000)


def get_fw_uuid(context):
    # type: (BundleContext) -> str
    """
    Returns the framework UUID

    :param context: The bundle context
    :return: The framework UUID
    """
    return str(context.get_property(constants.OSGI_FRAMEWORK_UUID))


def get_matching_interfaces(object_class, exported_intfs):
    # type: (List[str], Optional[List[str]]) -> Optional[List[str]]
    """
    Returns the list of interfaces matching the export property

    :param object_class: The specifications of the service
    :param exported_intfs: The declared exported interfaces
    :return: The list of declared exported interfaces
    """
    if object_class is None or exported_intfs is None:
        return None

    if isinstance(exported_intfs, str) and exported_intfs == "*":
        return object_class

    # after this exported_intfs will be list
    exported_intfs = get_string_plus_property_value(exported_intfs)
    if len(exported_intfs) == 1 and exported_intfs[0] == "*":
        return object_class

    return exported_intfs


def get_prop_value(name, props, default=None):
    # type: (str, Dict[str, Any], Any) -> Any
    """
    Returns the value of a property or the default one

    :param name: Name of a property
    :param props: Dictionary of properties
    :param default: Default value
    :return: The value of the property or the default one
    """
    if not props:
        return default

    try:
        return props[name]
    except KeyError:
        return default


def set_prop_if_null(name, props, if_null):
    # type: (str, Dict[str, Any], Any) -> None
    """
    Updates the value of a property if the previous one was None

    :param name: Name of the property
    :param props: Dictionary of properties
    :param if_null: Value to insert if the previous was None
    """
    value = get_prop_value(name, props)
    if value is None:
        props[name] = if_null


def get_string_plus_property_value(value):
    # type: (Any) -> Optional[List[str]]
    """
    Converts a string or list of string into a list of strings

    :param value: A string or a list of strings
    :return: A list of strings or None
    """
    if value:
        if isinstance(value, str):
            return [value]
        if isinstance(value, list):
            return value
        if isinstance(value, tuple):
            return list(value)

    return None


def convert_string_plus_value(values):
    """
    Normalizes a list of string

    :param values: A list of strings
    :return: None, a single string or a list of strings
    """
    if not values:
        return None

    if len(values) == 1:
        return values[0]

    return values


def parse_string_plus_value(value):
    # type: (str) -> List[str]
    """
    Parses a comma-separated value

    :param value: A string representation of a list
    :return: A list of strings
    """
    return value.split(",")


def get_string_plus_property(name, props, default=None):
    # type: (str, Dict[str, Any], Optional[Any]) -> Any
    """
    Returns the value of the given property or the default value

    :param name: A property name
    :param props: A dictionary of properties
    :param default: Value to return if the property doesn't exist
    :return: The property value or the default one
    """
    val = get_string_plus_property_value(get_prop_value(name, props, default))
    return default if val is None else val


def get_current_time_millis():
    # type: () -> int
    """
    Gets the current time stamp in milliseconds

    :return: The current time stamp
    """
    return int(time.time() * 1000)


def get_exported_interfaces(svc_ref, overriding_props=None):
    # type: (ServiceReference, Optional[Dict[str, Any]]) -> Optional[List[str]]
    """
    Looks for the interfaces exported by a service

    :param svc_ref: Service reference
    :param overriding_props: Properties overriding service ones
    :return: The list of exported interfaces
    """
    # first check overriding_props for service.exported.interfaces
    exported_intfs = get_prop_value(
        SERVICE_EXPORTED_INTERFACES, overriding_props
    )
    # then check svc_ref property
    if not exported_intfs:
        exported_intfs = svc_ref.get_property(SERVICE_EXPORTED_INTERFACES)

    if not exported_intfs:
        return None

    return get_matching_interfaces(
        svc_ref.get_property(constants.OBJECTCLASS), exported_intfs
    )


def validate_exported_interfaces(object_class, exported_intfs):
    # type: (List[str], List[str]) -> bool
    """
    Validates that the exported interfaces are all provided by the service

    :param object_class: The specifications of a service
    :param exported_intfs: The exported specifications
    :return: True if the exported specifications are all provided by the service
    """
    if (
        not exported_intfs
        or not isinstance(exported_intfs, list)
        or not exported_intfs
    ):
        return False
    else:
        for exintf in exported_intfs:
            if exintf not in object_class:
                return False
    return True


def get_package_from_classname(class_name):
    # type: (str) -> Optional[str]
    """
    Returns the name of the package declaring the given class

    :param class_name: A full class name
    :return: The name of a package or None
    """
    try:
        return class_name[: class_name.rindex(".")]
    except ValueError:
        return None


def get_package_versions(intfs, props):
    # type: (List[str], Dict[str, Any]) -> List[Tuple[str, str]]
    """
    Gets the package version of interfaces

    :param intfs: A list of interfaces
    :param props: A dictionary containing endpoint package versions
    :return: A list of tuples (package name, version)
    """
    result = []
    for intf in intfs:
        pkg_name = get_package_from_classname(intf)
        if pkg_name:
            key = ENDPOINT_PACKAGE_VERSION_ + pkg_name
            val = props.get(key, None)
            if val:
                result.append((key, val))
    return result


_NEXT_RSID = 1
_NEXT_RSID_LOCK = threading.Lock()


def get_next_rsid():
    # type: () -> int
    """
    Gets the next RS ID and increments the counter

    :return: The next RS ID
    """
    global _NEXT_RSID
    with _NEXT_RSID_LOCK:
        new_rsid = _NEXT_RSID
        _NEXT_RSID += 1
        return new_rsid


def copy_ref_props(service_ref):
    # type: (ServiceReference) -> Dict[str, Any]
    """
    Copies the properties of a service reference

    :param service_ref: A service reference
    :return: A copy of properties of the service
    """
    return service_ref.get_properties().copy()


def merge_dicts(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def merge_overriding_props(service_ref, overriding_props):
    # type: (ServiceReference, Dict[str, Any]) -> Dict[str, Any]
    """
    Overrides the properties of the service with the given ones

    :param service_ref: A service reference
    :param overriding_props: Properties overriding the service ones
    :return: The merged properties dictionary
    """
    ref_props = copy_ref_props(service_ref)
    return merge_dicts(ref_props, overriding_props)


def get_rsa_props(
    object_class,
    exported_cfgs,
    remote_intents=None,
    ep_svc_id=None,
    fw_id=None,
    pkg_vers=None,
    service_intents=None,
):
    """
    Constructs a dictionary of RSA properties from the given arguments

    :param object_class: Service specifications
    :param exported_cfgs: Export configurations
    :param remote_intents: Supported remote intents
    :param ep_svc_id: Endpoint service ID
    :param fw_id: Remote Framework ID
    :param pkg_vers: Version number of the specification package
    :param service_intents: Service intents
    :return: A dictionary of properties
    """
    results = {}
    if not object_class:
        raise ArgumentError(
            "object_class", "object_class must be an [] of Strings"
        )
    results["objectClass"] = object_class
    if not exported_cfgs:
        raise ArgumentError(
            "exported_cfgs", "exported_cfgs must be an array of Strings"
        )
    results[REMOTE_CONFIGS_SUPPORTED] = exported_cfgs
    results[SERVICE_IMPORTED_CONFIGS] = exported_cfgs
    if remote_intents:
        results[REMOTE_INTENTS_SUPPORTED] = remote_intents
    if service_intents:
        results[SERVICE_INTENTS] = service_intents
    if not ep_svc_id:
        ep_svc_id = get_next_rsid()
    results[ENDPOINT_SERVICE_ID] = ep_svc_id
    results[SERVICE_ID] = ep_svc_id
    if not fw_id:
        # No framework ID means an error
        fw_id = "endpoint-in-error"
    results[ENDPOINT_FRAMEWORK_UUID] = fw_id
    if pkg_vers:
        if isinstance(pkg_vers, type(tuple())):
            pkg_vers = [pkg_vers]
        for pkg_ver in pkg_vers:
            results[pkg_ver[0]] = pkg_ver[1]
    results[ENDPOINT_ID] = create_uuid()
    results[SERVICE_IMPORTED] = "true"
    return results


def get_ecf_props(ep_id, ep_id_ns, rsvc_id=None, ep_ts=None):
    """
    Prepares the ECF properties

    :param ep_id: Endpoint ID
    :param ep_id_ns: Namespace of the Endpoint ID
    :param rsvc_id: Remote service ID
    :param ep_ts: Timestamp of the endpoint
    :return: A dictionary of ECF properties
    """
    results = {}
    if not ep_id:
        raise ArgumentError("ep_id", "ep_id must be a valid endpoint id")
    results[ECF_ENDPOINT_ID] = ep_id
    if not ep_id_ns:
        raise ArgumentError("ep_id_ns", "ep_id_ns must be a valid namespace")
    results[ECF_ENDPOINT_CONTAINERID_NAMESPACE] = ep_id_ns
    if not rsvc_id:
        rsvc_id = get_next_rsid()
    results[ECF_RSVC_ID] = rsvc_id
    if not ep_ts:
        ep_ts = time_since_epoch()
    results[ECF_ENDPOINT_TIMESTAMP] = ep_ts
    return results


def get_extra_props(props):
    # type: (Dict[str, Any]) -> Dict[str, Any]
    """
    Returns the extra properties, *i.e.* non-ECF, non-RSA properties

    :param props: A dictionary of properties
    :return: A filtered dictionary
    """
    return {
        key: value
        for key, value in props.items()
        if key not in ECFPROPNAMES
        and key not in RSA_PROP_NAMES
        and not key.startswith(ENDPOINT_PACKAGE_VERSION_)
    }


def get_edef_props(
    object_class,
    exported_cfgs,
    ep_namespace,
    ep_id,
    ecf_ep_id,
    ep_rsvc_id,
    ep_ts,
    remote_intents=None,
    fw_id=None,
    pkg_ver=None,
    service_intents=None,
):
    """
    Prepares the EDEF properties of an endpoint, merge of RSA and ECF
    properties
    """
    osgi_props = get_rsa_props(
        object_class,
        exported_cfgs,
        remote_intents,
        ep_rsvc_id,
        fw_id,
        pkg_ver,
        service_intents,
    )
    ecf_props = get_ecf_props(ecf_ep_id, ep_namespace, ep_rsvc_id, ep_ts)
    return merge_dicts(osgi_props, ecf_props)


def get_edef_props_error(object_class):
    # type: (List[str]) -> Dict[str, Any]
    """
    Returns the EDEF properties for an errorred endpoint
    """
    return get_edef_props(
        object_class,
        ERROR_IMPORTED_CONFIGS,
        ERROR_NAMESPACE,
        ERROR_EP_ID,
        ERROR_ECF_EP_ID,
        0,
        0,
    )


def get_dot_properties(prefix, props, remove_prefix):
    # type: (str, Dict[str, Any], bool) -> Dict[str, Any]
    """
    Gets the properties starting with the given prefix
    """
    result_props = {}
    if props:
        dot_keys = [x for x in props.keys() if x.startswith(prefix + ".")]
        for dot_key in dot_keys:
            if remove_prefix:
                new_key = dot_key[len(prefix) + 1 :]
            else:
                new_key = dot_key
            result_props[new_key] = props.get(dot_key)
    return result_props


def is_reserved_property(key):
    # type: (str) -> bool
    """
    Tests if the given property key is reserved

    Reserved keys are:
    * RSA property names
    * ECF property names
    * Property names starting with a dot (``.``)

    :param key: A property name
    :return: True if the property is reserved
    """
    return key in RSA_PROP_NAMES or key in ECFPROPNAMES or key.startswith(".")


def remove_from_props(props, keys):
    # type: (Dict[str, Any], List[str]) -> Dict[str, Any]
    """
    Removes in-place the given keys from the properties

    :param props: A dictionary of properties
    :param keys: The keys to remove
    :return: The given dictionary of properties
    """
    for key in keys:
        try:
            del props[key]
        except KeyError:
            pass
    return props


def copy_non_reserved(props, target):
    # type: (Dict[str, Any], Dict[str, Any]) -> Dict[str, Any]
    """
    Copies all properties with non-reserved names from ``props`` to ``target``

    :param props: A dictionary of properties
    :param target: Another dictionary
    :return: The target dictionary
    """
    target.update(
        {
            key: value
            for key, value in props.items()
            if not is_reserved_property(key)
        }
    )
    return target


def copy_non_ecf(props, target):
    # type: (Dict[str, Any], Dict[str, Any]) -> Dict[str, Any]
    """
    Copies non-ECF properties from ``props`` to ``target``

    :param props: An input dictionary
    :param target: The dictionary to copy non-ECF properties to
    :return: The ``target`` dictionary
    """
    target.update(
        {key: value for key, value in props.items() if key not in ECFPROPNAMES}
    )
    return target


def set_append(input_set, item):
    # type: (set, Any) -> set
    """
    Appends in-place the given item to the set.
    If the item is a list, all elements are added to the set.

    :param input_set: An existing set
    :param item: The item or list of items to add
    :return: The given set
    """
    if item:
        if isinstance(item, (list, tuple)):
            input_set.update(item)
        else:
            input_set.add(item)
    return input_set


def cid_to_string(cid):
    # type: (Tuple[str, str]) -> str
    """
    Converts the Container ID to a string

    :param cid: A Container ID tuple
    :return: The Container ID as a string
    """
    return cid[1]


def rsid_to_string(rsid):
    # type: (Tuple[Tuple[str, str], int]) -> str
    """
    Converts the RS ID tuple to a string

    :param rsid: An RS ID tuple
    :return: The RS ID as a string
    """
    return "{0}:{1}".format(cid_to_string(rsid[0]), rsid[1])


def prop_dot_suffix(prop_name, suffix=None):
    # type: (str, Optional[str]) -> str
    """
    Joins both strings with a dot (".")
    """
    if not suffix:
        suffix = ""
    return "{0}.{1}".format(prop_name, suffix)


# ------------------------------------------------------------------------------
# Exception classes


class SelectExporterError(Exception):
    """
    Error selecting exporter
    """

    pass


class SelectImporterError(Exception):
    """
    Error selecting importer
    """

    pass


class RemoteServiceError(Exception):
    """
    Generic RSA exception
    """

    pass



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
A bridge to publish and subscribe to EventAdmin events over the network using
MQTT

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import json
import logging

# Pelix
from pelix.ipopo.decorators import (
    ComponentFactory,
    Provides,
    Property,
    Validate,
    Invalidate,
    Requires,
)
from pelix.utilities import to_str
import pelix.constants as constants
import pelix.misc.mqtt_client
import pelix.services as services

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

DEFAULT_MQTT_TOPIC = "pelix/eventadmin"
""" Default MQTT topic to use to propagate events """

EVENT_PROP_SOURCE_UID = "pelix.eventadmin.mqtt.source"
""" UID of the framework that sent the event """

EVENT_PROP_STARTING_SLASH = "pelix.eventadmin.mqtt.start_slash"
""" Flag to indicate that the EventAdmin topic starts with a '/' """

# ------------------------------------------------------------------------------


@ComponentFactory(services.FACTORY_EVENT_ADMIN_MQTT)
@Provides(services.SERVICE_EVENT_HANDLER, "_controller")
@Requires("_event", services.SERVICE_EVENT_ADMIN)
@Property("_event_topics", services.PROP_EVENT_TOPICS, "*")
@Property("_host", "mqtt.host", "localhost")
@Property("_port", "mqtt.port", 1883)
@Property("_mqtt_topic", "mqtt.topic.prefix", DEFAULT_MQTT_TOPIC)
class MqttEventAdminBridge(object):
    """
    The EventAdmin MQTT bridge
    """

    def __init__(self):
        """
        Sets up the members
        """
        # MQTT configuration
        self._host = "localhost"
        self._port = 1883
        self._mqtt_topic = None

        # MQTT Client
        self._mqtt = None

        # EventAdmin
        self._event = None
        self._event_topics = None

        # EventHandler service controller
        self._controller = False

        # Framework UID
        self._framework_uid = None

    @Validate
    def _validate(self, context):
        """
        Component validated
        """
        # Store the framework UID
        self._framework_uid = context.get_property(constants.FRAMEWORK_UID)

        if not self._mqtt_topic:
            # No topic given, use the default one
            self._mqtt_topic = DEFAULT_MQTT_TOPIC

        if self._mqtt_topic[-1] == "/":
            # Remove trailing slash
            self._mqtt_topic = self._mqtt_topic[:-1]

        # Create the MQTT client
        self._mqtt = pelix.misc.mqtt_client.MqttClient()

        # Customize callbacks
        self._mqtt.on_connect = self.__on_connect
        self._mqtt.on_disconnect = self.__on_disconnect
        self._mqtt.on_message = self.__on_message

        # Do not provide the EventHandler service before being connected
        self._controller = False

        # Prepare the connection
        self._mqtt.connect(self._host, self._port)

    @Invalidate
    def _invalidate(self, _):
        """
        Component invalidated
        """
        # Disconnect from the server (this stops the loop)
        self._mqtt.disconnect()

        # Clean up
        self._framework_uid = None
        self._mqtt = None

    def _make_topic(self, suffix):
        """
        Prepares a MQTT topic with the given suffix

        :param suffix: Suffix to the MQTT bridge topic
        :return: A MQTT topic
        """
        return "{0}/{1}".format(self._mqtt_topic, suffix)

    def __on_connect(self, client, result_code):
        """
        Client connected to the server
        """
        if not result_code:
            # Connection is OK, subscribe to the topic
            client.subscribe(self._make_topic("#"))

            # Provide the service
            self._controller = True

    def __on_disconnect(self, client, result_code):
        # pylint: disable=W0613
        """
        Client has been disconnected from the server
        """
        # Disconnected: stop providing the service
        self._controller = False

    def __on_message(self, client, msg):
        # pylint: disable=W0613
        """
        A message has been received from a server

        :param client: Client that received the message
        :param msg: A MQTTMessage bean
        """
        try:
            self.handle_mqtt_message(msg.topic, msg.payload)
        except Exception as ex:
            _logger.exception(
                "Error handling an MQTT EventAdmin message: %s", ex
            )

    def handle_event(self, topic, properties):
        """
        An EventAdmin event has been received
        """
        # Check that the event wasn't sent by us
        if EVENT_PROP_SOURCE_UID in properties:
            # A bridge posted this event, ignore it
            return
        elif services.EVENT_PROP_PROPAGATE not in properties:
            # Propagation flag is not set, ignore
            _logger.warning("No propagate")
            return

        # Remove starting '/' in the event, and set up the flag
        if topic[0] == "/":
            topic = topic[1:]
            properties[EVENT_PROP_STARTING_SLASH] = True

        # Prepare MQTT data
        mqtt_topic = self._make_topic(topic)
        payload = json.dumps(properties)

        # Publish the event to everybody, with QOS 2
        self._mqtt.publish(mqtt_topic, payload, qos=2)

    def handle_mqtt_message(self, mqtt_topic, payload):
        """
        An MQTT message has been received

        :param mqtt_topic: MQTT message topic
        :param payload: Payload of the message
        """
        # +1 to ignore the joining slash (prefix => prefix/)
        evt_topic = mqtt_topic[len(self._mqtt_topic) + 1 :]
        if not evt_topic:
            # Empty EventAdmin topic
            _logger.debug("Empty EventAdmin topic: %s", mqtt_topic)
            return

        try:
            # Ensure that the payload is a string
            payload = to_str(payload)

            # Parse the event payload
            properties = json.loads(payload)
        except ValueError as ex:
            # Oops...
            _logger.error("Error parsing the payload of %s: %s", evt_topic, ex)
            return

        # Check framework UID of the sender
        try:
            sender_uid = to_str(properties[services.EVENT_PROP_FRAMEWORK_UID])
            if sender_uid == self._framework_uid:
                # Loop back
                return

            # Set up source UID as an extra property
            properties[EVENT_PROP_SOURCE_UID] = sender_uid
        except KeyError:
            # Not sent by us... continue
            pass

        # Update the topic if necessary
        if properties.pop(EVENT_PROP_STARTING_SLASH, False):
            # Topic has a starting '/'
            evt_topic = "/{0}".format(evt_topic)

        # Post the event
        self._event.post(evt_topic, properties)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix basic HTTP service bundle.

Provides an implementation of the Pelix HTTP service based on the standard
Python library.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import socket
import threading
import traceback

# Basic HTTP server
try:
    # Python 3
    # pylint: disable=F0401,E0611
    from http.server import HTTPServer
    from http.server import BaseHTTPRequestHandler
    from socketserver import ThreadingMixIn, TCPServer
except ImportError:
    # Python 2 or IronPython
    # pylint: disable=F0401
    from BaseHTTPServer import HTTPServer
    from BaseHTTPServer import BaseHTTPRequestHandler
    from SocketServer import ThreadingMixIn, TCPServer

# iPOPO
from pelix.ipopo.decorators import (
    ComponentFactory,
    Provides,
    Requires,
    Validate,
    Invalidate,
    Property,
    HiddenProperty,
    BindField,
    UpdateField,
    UnbindField,
)
import pelix.ipopo.constants as constants
import pelix.ipv6utils
import pelix.utilities as utilities
import pelix.misc.ssl_wrap as ssl_wrap
import pelix.remote

# HTTP service constants
import pelix.http as http

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

HTTP_SERVICE_EXTRA = "http.extra"
""" HTTP service extra properties (dictionary) """

DEFAULT_BIND_ADDRESS = "0.0.0.0"
""" By default, bind to all IPv4 interfaces """

LOCALHOST_ADDRESS = "127.0.0.1"
"""
Local address, if None is given as binding address, instead of the default one
"""

# ------------------------------------------------------------------------------


class _HTTPServletRequest(http.AbstractHTTPServletRequest):
    """
    HTTP Servlet request helper
    """

    def __init__(self, request_handler, prefix):
        """
        Sets up the request helper

        :param request_handler: The basic request handler
        :param prefix: Teh path to the servlet root
        """
        self._handler = request_handler
        self._prefix = prefix

        # Compute the sub path
        self._sub_path = self._handler.path[len(prefix) :]
        if not self._sub_path.startswith("/"):
            self._sub_path = "/{0}".format(self._sub_path)

        while "//" in self._sub_path:
            self._sub_path = self._sub_path.replace("//", "/")

    def get_command(self):
        """
        Returns the HTTP verb (GET, POST, ...) used for the request
        """
        return self._handler.command

    def get_client_address(self):
        """
        Retrieves the address of the client

        :return: A (host, port) tuple
        """
        return self._handler.client_address

    def get_header(self, name, default=None):
        """
        Retrieves the value of a header
        """
        return self._handler.headers.get(name, default)

    def get_headers(self):
        """
        Retrieves all headers
        """
        return self._handler.headers

    def get_path(self):
        """
        Retrieves the request full path
        """
        return self._handler.path

    def get_prefix_path(self):
        """
        Returns the path to the servlet root

        :return: A request path (string)
        """
        return self._prefix

    def get_sub_path(self):
        """
        Returns the servlet-relative path, i.e. after the prefix

        :return: A request path (string)
        """
        return self._sub_path

    def get_rfile(self):
        """
        Retrieves the input as a file stream
        """
        return self._handler.rfile


class _HTTPServletResponse(http.AbstractHTTPServletResponse):
    """
    HTTP Servlet response helper
    """

    def __init__(self, request_handler):
        """
        Sets up the response helper

        :param request_handler: The basic request handler
        """
        self._handler = request_handler
        self._headers = {}

    def set_response(self, code, message=None):
        """
        Sets the response line.
        This method should be the first called when sending an answer.

        :param code: HTTP result code
        :param message: Associated message
        """
        self._handler.send_response(code, message)

    def set_header(self, name, value):
        """
        Sets the value of a header.
        This method should not be called after ``end_headers()``.

        :param name: Header name
        :param value: Header value
        """
        self._headers[name.lower()] = value

    def is_header_set(self, name):
        """
        Checks if the given header has already been set

        :param name: Header name
        :return: True if it has already been set
        """
        return name.lower() in self._headers

    def end_headers(self):
        """
        Ends the headers part
        """
        # Send them all at once
        for name, value in self._headers.items():
            self._handler.send_header(name, value)

        self._handler.end_headers()

    def get_wfile(self):
        """
        Retrieves the output as a file stream.
        ``end_headers()`` should have been called before, except if you want
        to write your own headers.

        :return: The output file-like object
        """
        return self._handler.wfile

    def write(self, data):
        """
        Writes the given data.
        ``end_headers()`` should have been called before, except if you want
        to write your own headers.

        :param data: Data to be written
        """
        self._handler.wfile.write(data)


# ------------------------------------------------------------------------------


class _RequestHandler(BaseHTTPRequestHandler, object):
    """
    Basic HTTP server request handler
    """

    # Override the default HTTP version
    default_request_version = "HTTP/1.0"

    def __init__(self, http_svc, *args, **kwargs):
        """
        Sets up the request handler (called for each request)

        :param http_svc: The associated HTTP service
        """
        self._service = http_svc

        # This calls the do_* methods
        BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def __getattr__(self, name):
        """
        Retrieves the do_* in the servlet corresponding to the request path.
        If the name is not a "do_*", returns the normal result of __getattr__.

        :param name: Name of the attribute
        :return: The found attribute
        :raise AttributeError: Attribute not found
        """
        if not name.startswith("do_"):
            # Not a request handling
            return object.__getattribute__(self, name)

        # Remove the query part and double-slashes in the request path
        parsed_path = self.path.split("?", 1)[0].replace("//", "/")

        # Get the corresponding servlet
        found_servlet = self._service.get_servlet(parsed_path)
        if found_servlet is not None:
            servlet, _, prefix = found_servlet
            if hasattr(servlet, name):
                # Prepare the helpers
                request = _HTTPServletRequest(self, prefix)
                response = _HTTPServletResponse(self)

                # Create a wrapper to pass the handler to the servlet
                def wrapper():
                    """
                    Wrapped servlet call
                    """
                    try:
                        # Handle the request
                        return getattr(servlet, name)(request, response)
                    except:
                        # Send a 500 error page on error
                        return self.send_exception(response)

                # Return it
                return wrapper

        # Return the super implementation if needed
        return self.send_no_servlet_response

    def log_error(self, message, *args, **kwargs):
        # pylint: disable=W0221
        """
        Log server error
        """
        self._service.log(logging.ERROR, message, *args, **kwargs)

    def log_request(self, code="-", size="-"):
        """
        Logs a request to the server
        """
        self._service.log(logging.DEBUG, '"%s" %s', self.requestline, code)

    def send_no_servlet_response(self):
        """
        Default response sent when no servlet is found for the requested path
        """
        # Use the helper to send the error page
        response = _HTTPServletResponse(self)
        response.send_content(404, self._service.make_not_found_page(self.path))

    def send_exception(self, response):
        """
        Sends an exception page with a 500 error code.
        Must be called from inside the exception handling block.

        :param response: The response handler
        """
        # Get a formatted stack trace
        stack = traceback.format_exc()

        # Log the error
        self.log_error(
            "Error handling request upon: %s\n%s\n", self.path, stack
        )

        # Send the page
        response.send_content(
            500, self._service.make_exception_page(self.path, stack)
        )


# ------------------------------------------------------------------------------


class _HttpServerFamily(ThreadingMixIn, HTTPServer):
    """
    A small modification to have a threaded HTTP Server with a custom address
    family

    Inspired from:
    http://www.arcfn.com/2011/02/ipv6-web-serving-with-arc-or-python.html
    """

    def __init__(
        self,
        server_address,
        request_handler_class,
        request_queue_size=5,
        logger=None,
    ):
        """
        Proxy constructor

        :param server_address: The server address
        :param request_handler_class: The request handler class
        :param request_queue_size: The size of the request queue
                                   (clients waiting for treatment)
        :param logger: An optional logger, in case of ignored error
        """
        # Determine the address family
        addr_info = socket.getaddrinfo(
            server_address[0], server_address[1], 0, 0, socket.SOL_TCP
        )

        # Change the address family before the socket is created
        # Get the family of the first possibility
        self.address_family = addr_info[0][0]

        # Set the queue size
        self.request_queue_size = request_queue_size

        # Set up the server, socket, ... but do not bind immediately
        HTTPServer.__init__(self, server_address, request_handler_class, False)
        self.server_name = server_address[0]
        self.server_port = server_address[1]

        if self.address_family == socket.AF_INET6:
            # Explicitly ask to be accessible both by IPv4 and IPv6
            try:
                pelix.ipv6utils.set_double_stack(self.socket)
            except AttributeError as ex:
                if logger is not None:
                    logger.exception("System misses IPv6 constant: %s", ex)
            except socket.error as ex:
                if logger is not None:
                    logger.exception(
                        "Error setting up IPv6 double stack: %s", ex
                    )

        # Bind & accept
        self.server_bind()
        self.server_activate()

    def server_bind(self):
        """
        Override server_bind to store the server name, even in IronPython.

        See https://ironpython.codeplex.com/workitem/29477
        """
        TCPServer.server_bind(self)
        host, port = self.socket.getsockname()[:2]
        self.server_port = port
        try:
            self.server_name = socket.getfqdn(host)
        except ValueError:
            # Use the local host name in case of error, like CPython does
            self.server_name = socket.gethostname()

    def process_request(self, request, client_address):
        """
        Starts a new thread to process the request, adding the client address
        in its name.
        """
        thread = threading.Thread(
            name="HttpService-{0}-Client-{1}".format(
                self.server_port, client_address
            ),
            target=self.process_request_thread,
            args=(request, client_address),
        )
        thread.daemon = self.daemon_threads
        thread.start()


# ------------------------------------------------------------------------------


@ComponentFactory(http.FACTORY_HTTP_BASIC)
@Provides(http.HTTP_SERVICE)
@Requires("_servlets_services", http.HTTP_SERVLET, True, True)
@Requires("_error_handler", http.HTTP_ERROR_PAGES, optional=True)
@Property("_address", http.HTTP_SERVICE_ADDRESS, DEFAULT_BIND_ADDRESS)
@Property("_port", http.HTTP_SERVICE_PORT, 8080)
@Property("_uses_ssl", http.HTTP_USES_SSL, False)
@Property("_cert_file", http.HTTPS_CERT_FILE, None)
@Property("_key_file", http.HTTPS_KEY_FILE, None)
@HiddenProperty("_key_password", http.HTTPS_KEY_PASSWORD, None)
@Property("_extra", HTTP_SERVICE_EXTRA, None)
@Property("_instance_name", constants.IPOPO_INSTANCE_NAME)
@Property("_logger_name", "pelix.http.logger.name", "")
@Property("_logger_level", "pelix.http.logger.level", None)
@Property("_request_queue_size", "pelix.http.request_queue_size", 100)
class HttpService(object):
    """
    Basic HTTP service component
    """

    def __init__(self):
        # Properties
        self._address = "0.0.0.0"
        self._port = 8080
        self._uses_ssl = False
        self._extra = None
        self._instance_name = None
        self._logger_name = None
        self._logger_level = None
        self._request_queue_size = 5

        # SSL Parameters
        self._cert_file = None
        self._key_file = None
        self._key_password = None

        # Validation flag
        self._validated = False

        # The logger
        self._logger = None

        # Servlets registry lock
        self._lock = threading.Lock()

        # Path -> (servlet, parameters)
        self._servlets = {}

        # Fields injected by iPOPO
        self._servlets_services = None
        self._error_handler = None

        # Servlet -> ServiceReference
        self._servlets_refs = {}
        self._binding_lock = threading.Lock()

        # Server control
        self._server = None
        self._thread = None

    def __str__(self):
        """
        String representation of the instance
        """
        return "BasicHttpService({0}, {1:d})".format(self._address, self._port)

    def __safe_callback(self, instance, method, *args, **kwargs):
        """
        Safely calls the given method in the given instance.
        Returns True on method absence.
        Returns False on error.
        Returns the method result if found.

        :param instance: The instance to call
        :param method: The method to call in the instance
        :return: The method result or True on method absence or False on error
        """
        # Call back the method
        if instance is None:
            # Consider invalidity as a failure
            return False

        try:
            callback = getattr(instance, method)
        except AttributeError:
            # Consider absence as a success
            return True

        try:
            result = callback(*args, **kwargs)
            if result is None:
                # Special case: consider None as success
                return True

            return result

        except Exception as ex:
            self.log_exception("Error calling back an instance: %s", ex)

        return False

    def __register_servlet_service(self, service, service_reference):
        """
        Registers a servlet according to its service properties

        :param service: A servlet service
        :param service_reference: The associated ServiceReference
        """
        # Servlet bound
        paths = service_reference.get_property(http.HTTP_SERVLET_PATH)
        if utilities.is_string(paths):
            # Register the servlet to a single path
            self.register_servlet(paths, service)
        elif isinstance(paths, (list, tuple)):
            # Register the servlet to multiple paths
            for path in paths:
                self.register_servlet(path, service)

    @BindField("_servlets_services")
    def _bind(self, _, service, service_reference):
        """
        Called by iPOPO when a service is bound
        """
        # Ignore imported services
        if self.__is_imported(service_reference):
            return
        with self._binding_lock:
            self._servlets_refs[service] = service_reference

            if self._validated:
                # We've been validated, register the service
                self.__register_servlet_service(service, service_reference)

    @UpdateField("_servlets_services")
    def _update(self, _, service, service_reference, old_properties):
        """
        Called by iPOPO when the properties of a service have been updated
        """
        # Ignore imported services
        if self.__is_imported(service_reference):
            return
        # Check if the property concerns the registration
        old_path = old_properties.get(http.HTTP_SERVLET_PATH)
        new_path = service_reference.get_property(http.HTTP_SERVLET_PATH)
        if old_path == new_path:
            # Nothing to do
            return

        with self._binding_lock:
            # Unregister the previous paths
            self.unregister(None, service)

            if self._validated:
                # Register the service with its new properties
                self.__register_servlet_service(service, service_reference)

    @UnbindField("_servlets_services")
    def _unbind(self, _, service, service_reference):
        """
        Called by iPOPO when a service is gone
        """
        # Ignore imported services
        if self.__is_imported(service_reference):
            return
        with self._binding_lock:
            # Servlet gone: unregister all paths associated to this servlet
            self.unregister(None, service)

            # Remove the service reference
            del self._servlets_refs[service]

    def get_access(self):
        """
        Retrieves the (address, port) tuple to access the server
        """
        sock_info = self._server.socket.getsockname()

        # Only keep the address and the port information
        return sock_info[0], sock_info[1]

    @staticmethod
    def get_hostname():
        """
        Retrieves the server host name

        :return: The server host name
        """
        return socket.gethostname()

    def is_https(self):
        """
        Returns True if this is an HTTPS server

        :return: True if this server uses SSL
        """
        return self._uses_ssl

    def get_registered_paths(self):
        """
        Returns the paths registered by servlets

        :return: The paths registered by servlets (sorted list)
        """
        return sorted(self._servlets)

    def get_servlet(self, path):
        """
        Retrieves the servlet matching the given path and its parameters.
        Returns None if no servlet matches the given path.

        :param path: A request URI
        :return: A tuple (servlet, parameters, prefix) or None
        """
        if not path or path[0] != "/":
            # No path, nothing to return
            return None

        # Use lower case for comparison
        path = path.lower()

        if path[-1] != "/":
            # Add a trailing slash
            path += "/"

        with self._lock:
            longest_match = ""
            longest_match_len = 0
            for servlet_path in self._servlets:
                tested_path = servlet_path
                if tested_path[-1] != "/":
                    # Add a trailing slash
                    tested_path += "/"

                if (
                    path.startswith(tested_path)
                    and len(servlet_path) > longest_match_len
                ):
                    # Found a corresponding servlet
                    # which is deeper than the previous one
                    longest_match = servlet_path
                    longest_match_len = len(servlet_path)

            # Return the found servlet
            if not longest_match:
                # No match found
                return None

            # Retrieve the stored information
            return tuple(self._servlets[longest_match]) + (longest_match,)

    def make_not_found_page(self, path):
        """
        Prepares a "page not found" page for a 404 error

        :param path: Request path
        :return: A HTML page
        """
        page = None
        if self._error_handler is not None:
            page = self._error_handler.make_not_found_page(path)

        if not page:
            page = """<html>
<head>
<title>404 - Page not found</title>
</head>
<body>
<h1>Page not found</h1>
<p>No servlet is associated to this path:</p>
<pre>{0}</pre>
<h2>Registered paths:</h2>
{1}
</body>
</html>""".format(
                path, http.make_html_list(self.get_registered_paths())
            )

        return page

    def make_exception_page(self, path, stack):
        """
        Prepares a page printing an exception stack trace in a 500 error

        :param path: Request path
        :param stack: Exception stack trace
        :return: A HTML page
        """
        page = None
        if self._error_handler is not None:
            page = self._error_handler.make_exception_page(path, stack)

        if not page:
            page = """<html>
<head>
<title>500 - Internal Server Error</title>
</head>
<body>
<h1>Internal Server Error</h1>
<p>Error handling request upon: {0}</p>
<pre>
{1}
</pre>
</body>
</html>""".format(
                path, stack
            )

        return page

    def register_servlet(self, path, servlet, parameters=None):
        """
        Registers a servlet

        :param path: Path handled by this servlet
        :param servlet: The servlet instance
        :param parameters: The parameters associated to this path
        :return: True if the servlet has been registered, False if it refused
                 the binding.
        :raise ValueError: Invalid path or handler
        """
        if servlet is None:
            raise ValueError("Invalid servlet instance")

        if not path or path[0] != "/":
            raise ValueError(
                "Invalid path given to register the servlet: {0}".format(path)
            )

        # Use lower-case paths
        path = path.lower()

        # Prepare the parameters
        if parameters is None:
            parameters = {}

        with self._lock:
            if path in self._servlets:
                # Already registered path
                if self._servlets[path][0] is servlet:
                    # Double-registration: Nothing to do
                    return True
                else:
                    # Path is already taken by another servlet
                    already_taken = True
            else:
                # Path is available
                already_taken = False

            # Add server information in parameters
            parameters[http.PARAM_ADDRESS] = self._address
            parameters[http.PARAM_PORT] = self._port
            parameters[http.PARAM_HTTPS] = self._uses_ssl
            parameters[http.PARAM_NAME] = self._instance_name
            parameters[http.PARAM_EXTRA] = self._extra.copy()

            # The servlet might refuse to be bound to this server
            if not self.__safe_callback(
                servlet, "accept_binding", path, parameters
            ):
                # Server refused: stop right there
                # => No need to raise the "already taken path" exception
                return False

            if already_taken:
                # The path is already taken by another servlet
                raise ValueError(
                    "A servlet is already registered on {0}".format(path)
                )

            # Tell the servlet it can be bound to the path
            if self.__safe_callback(servlet, "bound_to", path, parameters):
                # Store the servlet
                self._servlets[path] = (servlet, parameters)
                return True

            # The servlet refused the binding
            return False

    def unregister(self, path, servlet=None):
        """
        Unregisters the servlet for the given path

        :param path: The path to a servlet
        :param servlet: If given, unregisters all the paths handled by this
                        servlet

        :return: True if at least one path as been unregistered, else False
        """
        if servlet is not None:
            with self._lock:
                # Unregister all paths for this servlet
                paths = [
                    servlet_path
                    for (servlet_path, servlet_info) in self._servlets.items()
                    if servlet_info[0] == servlet
                ]

            result = False
            for servlet_path in paths:
                result |= self.unregister(servlet_path)

            return result
        else:
            if not path:
                # Invalid path
                return False

            # Always use lower case to compare paths
            path = path.lower()

            with self._lock:
                # Notify the servlet
                servlet_info = self._servlets.get(path)
                if servlet_info is None:
                    # Unknown path
                    return False

                self.__safe_callback(
                    servlet_info[0], "unbound_from", path, servlet_info[1]
                )

                # Remove the servlet
                del self._servlets[path]
                return True

    def log(self, level, message, *args, **kwargs):
        """
        Logs the given message

        :param level: Log entry level
        :param message: Log message (Python logging format)
        """
        if self._logger is not None:
            # Log the message
            self._logger.log(level, message, *args, **kwargs)

    def log_exception(self, message, *args, **kwargs):
        """
        Logs an exception

        :param message: Log message (Python logging format)
        """
        if self._logger is not None:
            # Log the exception
            self._logger.exception(message, *args, **kwargs)

    @Validate
    def validate(self, _):
        """
        Component validation
        """
        # Check if we'll use an SSL connection
        self._uses_ssl = self._cert_file is not None

        if not self._address:
            # No address given, use the localhost address
            self._address = LOCALHOST_ADDRESS

        if self._port is None:
            # Random port
            self._port = 0
        else:
            # Ensure we have an integer
            self._port = int(self._port)
            if self._port < 0:
                # Random port
                self._port = 0

        # Normalize the request queue size
        try:
            self._request_queue_size = int(self._request_queue_size)
        except (ValueError, TypeError):
            self._request_queue_size = 5

        if self._request_queue_size <= 0:
            self._request_queue_size = 5

        # Normalize the extra properties
        if not isinstance(self._extra, dict):
            self._extra = {}

        # Set up the logger
        if self._logger_name is not None:
            if not self._logger:
                # Empty name, use the instance name
                self._logger_name = self._instance_name

            self._logger = logging.getLogger(self._logger_name)

            if self._logger_level is None:
                self._logger.level = logging.INFO
            else:
                self._logger.level = int(self._logger_level)

        self.log(
            logging.INFO,
            "Starting HTTP%s server: [%s]:%d ...",
            "S" if self._uses_ssl else "",
            self._address,
            self._port,
        )

        # Create the server
        self._server = _HttpServerFamily(
            (self._address, self._port),
            lambda *x: _RequestHandler(self, *x),
            self._request_queue_size,
            self._logger,
        )

        if self._uses_ssl:
            # Activate HTTPS if required
            self._server.socket = ssl_wrap.wrap_socket(
                self._server.socket,
                self._cert_file,
                self._key_file,
                self._key_password,
            )

        # Property update (if port was 0)
        self._port = self._server.server_port

        # Run it in a separate thread
        self._thread = threading.Thread(
            target=self._server.serve_forever,
            name="HttpService-{0}-Server".format(self._port),
        )
        self._thread.daemon = True
        self._thread.start()

        with self._binding_lock:
            # Set the validation flag up, once the server is ready
            self._validated = True

            # Register bound servlets
            for service, svc_ref in self._servlets_refs.items():
                self.__register_servlet_service(service, svc_ref)

        self.log(
            logging.INFO,
            "HTTP%s server started: [%s]:%d",
            "S" if self._uses_ssl else "",
            self._address,
            self._port,
        )

    @Invalidate
    def invalidate(self, _):
        """
        Component invalidation
        """
        with self._binding_lock:
            # Refuse new registrations
            self._validated = False

            # Unregister servlets (to call unbound_from...)
            for service in self._servlets_refs:
                self.unregister(None, service)

        self.log(
            logging.INFO,
            "Shutting down HTTP server: [%s]:%d ...",
            self._address,
            self._port,
        )

        # Shutdown server (if active)
        if self._server is not None:
            self._server.shutdown()

            # Wait for the thread to stop...
            self.log(
                logging.INFO,
                "Waiting HTTP server ([%s]:%d) thread to stop...",
                self._address,
                self._port,
            )
            self._thread.join(2)

            # Close the server
            self._server.server_close()

        self.log(
            logging.INFO,
            "HTTP server down: [%s]:%d ...",
            self._address,
            self._port,
        )

        # Clean up
        self._servlets.clear()
        self._thread = None
        self._server = None
        self._logger = None

    @staticmethod
    def __is_imported(service_reference):
        """
        Tests if the given service has been imported by Remote Services

        :param service_reference: The reference of the service to check
        :return: True if the service is flagged as imported
        """
        return service_reference.get_property(pelix.remote.PROP_IMPORTED)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Defines the iPOPO decorators classes to manipulate component factory classes

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import inspect
import logging
import sys
import threading
import types

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, Callable
except ImportError:
    pass

# Pelix modules
from pelix.utilities import is_string, to_iterable, get_method_arguments
from pelix.ipopo.contexts import FactoryContext, Requirement
import pelix.ipopo.constants as constants

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

# Prepare the module logger
_logger = logging.getLogger("ipopo.decorators")

# ------------------------------------------------------------------------------


def is_from_parent(cls, attribute_name, value=None):
    # type: (type, str, bool) -> bool
    """
    Tests if the current attribute value is shared by a parent of the given
    class.

    Returns None if the attribute value is None.

    :param cls: Child class with the requested attribute
    :param attribute_name: Name of the attribute to be tested
    :param value: The exact value in the child class (optional)
    :return: True if the attribute value is shared with a parent class
    """
    if value is None:
        try:
            # Get the current value
            value = getattr(cls, attribute_name)
        except AttributeError:
            # No need to go further: the attribute does not exist
            return False

    for base in cls.__bases__:
        # Look for the value in each parent class
        try:
            return getattr(base, attribute_name) is value
        except AttributeError:
            pass

    # Attribute value not found in parent classes
    return False


def get_factory_context(cls):
    # type: (type) -> FactoryContext
    """
    Retrieves the factory context object associated to a factory. Creates it
    if needed

    :param cls: The factory class
    :return: The factory class context
    """
    context = getattr(cls, constants.IPOPO_FACTORY_CONTEXT, None)

    if context is None:
        # Class not yet manipulated
        context = FactoryContext()
    elif is_from_parent(cls, constants.IPOPO_FACTORY_CONTEXT):
        # Create a copy the context
        context = context.copy(True)
        # * Manipulation has not been applied yet
        context.completed = False
    else:
        # Nothing special to do
        return context

    # Context has been created or copied, inject the new bean
    setattr(cls, constants.IPOPO_FACTORY_CONTEXT, context)
    return context


def get_method_description(method):
    # type: (Callable) -> str
    """
    Retrieves a description of the given method. If possible, the description
    contains the source file name and line.

    :param method: A method
    :return: A description of the method (at least its name)
    :raise AttributeError: Given object has no __name__ attribute
    """
    try:
        try:
            line_no = inspect.getsourcelines(method)[1]
        except IOError:
            # Error reading the source file
            line_no = -1

        return "'{method}' ({file}:{line})".format(
            method=method.__name__, file=inspect.getfile(method), line=line_no
        )
    except TypeError:
        # Method can't be inspected
        return "'{0}'".format(method.__name__)


def validate_method_arity(method, *needed_args):
    # type: (Callable, *str) -> None
    """
    Tests if the decorated method has a sufficient number of parameters.

    :param method: The method to be tested
    :param needed_args: The name (for description only) of the needed
                        arguments, without "self".
    :return: Nothing
    :raise TypeError: Invalid number of parameter
    """
    nb_needed_args = len(needed_args)

    # Test the number of parameters
    arg_spec = get_method_arguments(method)
    method_args = arg_spec.args

    try:
        # Remove the self argument when present
        if method_args[0] == "self":
            del method_args[0]
    except IndexError:
        pass

    nb_args = len(method_args)

    if arg_spec.varargs is not None:
        # Variable arguments
        if nb_args != 0:
            # Other arguments detected
            raise TypeError(
                "When using '*args', the decorated {0} method must only "
                "accept the 'self' argument".format(
                    get_method_description(method)
                )
            )
    elif arg_spec.keywords is not None:
        raise TypeError("Methods using '**kwargs' are not handled")
    elif nb_args != nb_needed_args:
        # "Normal" arguments
        raise TypeError(
            "The decorated method {0} must accept exactly {1} parameters: "
            "(self, {2})".format(
                get_method_description(method),
                nb_needed_args + 1,
                ", ".join(needed_args),
            )
        )


# ------------------------------------------------------------------------------


def _ipopo_setup_callback(cls, context):
    # type: (type, FactoryContext) -> None
    """
    Sets up the class _callback dictionary

    :param cls: The class to handle
    :param context: The factory class context
    """
    assert inspect.isclass(cls)
    assert isinstance(context, FactoryContext)

    if context.callbacks is not None:
        callbacks = context.callbacks.copy()
    else:
        callbacks = {}

    functions = inspect.getmembers(cls, inspect.isroutine)

    for _, func in functions:
        if not hasattr(func, constants.IPOPO_METHOD_CALLBACKS):
            # No attribute, get the next member
            continue

        method_callbacks = getattr(func, constants.IPOPO_METHOD_CALLBACKS)
        if not isinstance(method_callbacks, list):
            # Invalid content
            _logger.warning(
                "Invalid callback information %s in %s",
                constants.IPOPO_METHOD_CALLBACKS,
                get_method_description(func),
            )
            continue

        # Keeping it allows inheritance : by removing it, only the first
        # child will see the attribute -> Don't remove it

        # Store the call backs
        for _callback in method_callbacks:
            if _callback in callbacks and not is_from_parent(
                cls, callbacks[_callback].__name__, callbacks[_callback]
            ):
                _logger.warning(
                    "Redefining the callback %s in class '%s'.\n"
                    "\tPrevious callback : %s\n"
                    "\tNew callback : %s",
                    _callback,
                    cls.__name__,
                    get_method_description(callbacks[_callback]),
                    get_method_description(func),
                )

            callbacks[_callback] = func

    # Update the factory context
    context.callbacks.clear()
    context.callbacks.update(callbacks)


def _ipopo_setup_field_callback(cls, context):
    # type: (type, FactoryContext) -> None
    """
    Sets up the class _field_callback dictionary

    :param cls: The class to handle
    :param context: The factory class context
    """
    assert inspect.isclass(cls)
    assert isinstance(context, FactoryContext)

    if context.field_callbacks is not None:
        callbacks = context.field_callbacks.copy()
    else:
        callbacks = {}

    functions = inspect.getmembers(cls, inspect.isroutine)
    for name, func in functions:
        if not hasattr(func, constants.IPOPO_METHOD_FIELD_CALLBACKS):
            # No attribute, get the next member
            continue

        method_callbacks = getattr(func, constants.IPOPO_METHOD_FIELD_CALLBACKS)
        if not isinstance(method_callbacks, list):
            # Invalid content
            _logger.warning(
                "Invalid attribute %s in %s",
                constants.IPOPO_METHOD_FIELD_CALLBACKS,
                name,
            )
            continue

        # Keeping it allows inheritance : by removing it, only the first
        # child will see the attribute -> Don't remove it

        # Store the call backs
        for kind, field, if_valid in method_callbacks:
            fields_cbs = callbacks.setdefault(field, {})

            if kind in fields_cbs and not is_from_parent(
                cls, fields_cbs[kind][0].__name__
            ):
                _logger.warning(
                    "Redefining the callback %s in '%s'. "
                    "Previous callback : '%s' (%s). "
                    "New callback : %s",
                    kind,
                    name,
                    fields_cbs[kind][0].__name__,
                    fields_cbs[kind][0],
                    func,
                )

            fields_cbs[kind] = (func, if_valid)

    # Update the factory context
    context.field_callbacks.clear()
    context.field_callbacks.update(callbacks)


# ------------------------------------------------------------------------------


def _set_object_entry(obj, entry_name, value):
    # type: (Any, str, Any) -> None
    """
    Sets the given value to the given attribute in the given object.

    :param obj: The object that contains the list
    :param entry_name: The name of the member in *obj*
    :param value: The value to set
    """
    setattr(obj, entry_name, value)


def _append_object_entry(obj, list_name, entry):
    # type: (Any, str, Any) -> None
    """
    Appends the given entry in the given object list.
    Creates the list field if needed.

    :param obj: The object that contains the list
    :param list_name: The name of the list member in *obj*
    :param entry: The entry to be added to the list
    :raise ValueError: Invalid attribute content
    """
    # Get the list
    obj_list = getattr(obj, list_name, None)
    if obj_list is None:
        # We'll have to create it
        obj_list = []
        setattr(obj, list_name, obj_list)

    assert isinstance(obj_list, list)

    # Set up the property, if needed
    if entry not in obj_list:
        obj_list.append(entry)


# ------------------------------------------------------------------------------


class Holder(object):
    # pylint: disable=R0903
    """
    Simple class that holds a value
    """

    def __init__(self, value):
        """
        Sets up the holder instance
        """
        self.value = value


def _ipopo_class_field_property(name, value, methods_prefix):
    # type: (str, Any, str) -> property
    """
    Sets up an iPOPO field property, using Python property() capabilities

    :param name: The property name
    :param value: The property default value
    :param methods_prefix: The common prefix of the getter and setter injected
                           methods
    :return: A generated Python property()
    """
    # The property lock
    lock = threading.RLock()

    # Prepare the methods names
    getter_name = "{0}{1}".format(methods_prefix, constants.IPOPO_GETTER_SUFFIX)
    setter_name = "{0}{1}".format(methods_prefix, constants.IPOPO_SETTER_SUFFIX)

    local_holder = Holder(value)

    def get_value(self):
        """
        Retrieves the property value, from the iPOPO dictionaries
        """
        getter = getattr(self, getter_name, None)
        if getter is not None:
            # Use the component getter
            with lock:
                return getter(self, name)
        else:
            # Use the local holder
            return local_holder.value

    def set_value(self, new_value):
        """
        Sets the property value and trigger an update event

        :param new_value: The new property value
        """
        setter = getattr(self, setter_name, None)
        if setter is not None:
            # Use the component setter
            with lock:
                setter(self, name, new_value)
        else:
            # Change the local holder
            local_holder.value = new_value

    return property(get_value, set_value)


# ------------------------------------------------------------------------------


class Instantiate(object):
    # pylint: disable=R0903
    """
    This decorator tells iPOPO to instantiate a component instance from this
    factory as soon as its bundle is in **ACTIVE** state.

    The ``properties`` argument allows to override the default value given in
    the ``@Property`` and ``@HiddenProperty`` decorators.

    The properties are associated to the component instance but not added to
    it. This means that new (meta-) properties can be added to add information
    to the component (like the Remote Services export properties), but those
    won't be accessible directly by the component.
    Those extra properties will be visible in component's services properties
    and in the instance properties returned by the iPOPO
    ``get_instance_details()`` method, but no new field will be injected in the
    component instance.

    .. code-block:: python

        @ComponentFactory()
        @Property('_name', 'name', 'foo')
        @Instantiate('component-1')
        @Instantiate('component-2', {'name': 'bar'})
        class Foo(object):
            def call(self):
                # component-1 will print "foo" (default value)
                # component-2 will print "bar"
                print("My name property is:", self._name)

        @ComponentFactory()
        @Provides("thermometer")
        @Property('_value', 'value', 1)
        @Instantiate('component-3', {'unit': 'K'})
        class Bar(object):
            def call(self):
                # We don't have access to the "unit" property value, but it
                # will be visible in the service properties
                print("My value is:", self._value)
    """

    def __init__(self, name, properties=None):
        """
        :param name: The name of the component instance (**mandatory**)
        :param properties: The initial properties of the instance as a
                           dictionary
        """
        if not is_string(name):
            raise TypeError("Instance name must be a string")

        if properties is not None and not isinstance(properties, dict):
            raise TypeError("Instance properties must be a dictionary or None")

        name = name.strip()
        if not name:
            raise ValueError("Invalid instance name '{0}'".format(name))

        self.__name = name
        self.__properties = properties

    def __call__(self, factory_class):
        """
        Sets up and registers the instances descriptions

        :param factory_class: The factory class to instantiate
        :return: The decorated factory class
        :raise TypeError: The given object is not a class
        """
        if not inspect.isclass(factory_class):
            raise TypeError(
                "@Instantiate can decorate only classes, "
                "not '{0}'".format(type(factory_class).__name__)
            )

        # Store the instance in the factory context
        context = get_factory_context(factory_class)
        try:
            context.add_instance(self.__name, self.__properties)

        except NameError:
            _logger.warning(
                "Component '%s' defined twice, new definition ignored",
                self.__name,
            )

        return factory_class


# ------------------------------------------------------------------------------


class ComponentFactory(object):
    # pylint: disable=R0903
    """
    Manipulates the component class according to a ``FactoryContext`` object
    filled by other decorators.

    This **must** be the last executed decorator, *i.e.* the one on top of
    others in the source code.

    If no factory name is given, it will be generated as ``ClassNameFactory``,
    *e.g.* a ``Foo`` class will have the factory name ``FooFactory``.

    Note that the name of the component factory is the only way to identify it
    in iPOPO. Therefore, it must be unique in a framework instance.

    .. warning:: The ``__init__()`` method of a component factory class must
                 not require any parameter.

    :Example:

    .. code-block:: python

        @ComponentFactory()
        class Foo(object):
            def __init__(self):
                pass

        @ComponentFactory('my-factory')
        class Bar(object):
           pass

        @ComponentFactory()
        class FooBar(object):
            def __init__(self, managed=True):
                # The argument has a default value: it can be instantiated by
                # iPOPO
                pass
    """

    def __init__(self, name=None, excluded=None):
        """
        :param name: Name of the component factory, used to identify it when
                     instantiating a component. This name must be unique in a
                     Pelix framework instance.
        :param excluded: List of IDs of handlers which configuration must
                         **not** be inherited from the parent class
        """
        self.__factory_name = name
        self.__excluded_inheritance = to_iterable(excluded)

    def __call__(self, factory_class):
        """
        Sets up and registers the factory class

        :param factory_class: The class to decorate
        :return: The decorated class
        :raise TypeError: The given object is not a class
        """
        if not inspect.isclass(factory_class):
            raise TypeError(
                "@ComponentFactory can decorate only classes, "
                "not '{0}'".format(type(factory_class).__name__)
            )

        # Get the factory context
        context = get_factory_context(factory_class)

        # Test if a manipulation has already been applied
        if not context.completed:
            # Set up the factory name
            if not self.__factory_name:
                self.__factory_name = factory_class.__name__ + "Factory"

            # Manipulate the class...

            # Update the factory context
            context.name = self.__factory_name
            context.inherit_handlers(self.__excluded_inheritance)
            context.is_singleton = False
            context.completed = True

            # Find callbacks
            _ipopo_setup_callback(factory_class, context)
            _ipopo_setup_field_callback(factory_class, context)

            # Store the factory context in its field
            setattr(factory_class, constants.IPOPO_FACTORY_CONTEXT, context)

            # Inject the properties getter and setter if needed
            if context.properties_fields:
                setattr(
                    factory_class,
                    constants.IPOPO_PROPERTY_PREFIX
                    + constants.IPOPO_GETTER_SUFFIX,
                    None,
                )
                setattr(
                    factory_class,
                    constants.IPOPO_PROPERTY_PREFIX
                    + constants.IPOPO_SETTER_SUFFIX,
                    None,
                )
        else:
            # Manipulation already applied: do nothing more
            _logger.error(
                "%s has already been manipulated with the name '%s'."
                " Keeping the old name.",
                get_method_description(factory_class),
                context.name,
            )

        return factory_class


class SingletonFactory(ComponentFactory):
    # pylint: disable=R0903
    """
    This decorator is a specialization of the :class:`~ComponentFactory`: it
    accepts the same arguments and follows the same rule, but it allows only
    one instance of component from this factory at a time.

    If the factory is instantiated while another already exist, a
    ``ValueError`` will be raised.

    .. code-block:: python

        @SingletonFactory()
        class Foo(object):
            def __init__(self):
                pass

        @SingletonFactory('my-factory')
        class Bar(object):
           pass
    """

    def __call__(self, factory_class):
        """
        Sets up and registers the factory class

        :param factory_class: The class to decorate
        :return: The decorated class
        :raise TypeError: The given object is not a class
        """
        # Manipulate the class
        factory_class = super(SingletonFactory, self).__call__(factory_class)

        # Set the singleton flag
        context = get_factory_context(factory_class)
        context.is_singleton = True
        return factory_class


# ------------------------------------------------------------------------------


class Property(object):
    # pylint: disable=R0903
    """
    The ``@Property`` decorator defines a component instance property.
    A property can be used to configure the component at instantiation time and
    to expose the state of a component.
    Note that component properties are exposed in the properties of the
    services it publishes with the :class:`Provides` decorator.

    If no initial value is given in the decorator, the value stored in the
    injected field in the ``__init__()`` method will be used.

    .. warning:: In Python 2, it is required that the component class inherits
                 ``object`` for properties to work.

    :Handler ID: :py:const:`pelix.ipopo.constants.HANDLER_PROPERTY`

    :Example:

    .. code-block:: python

        @ComponentFactory()
        @Property('_answer', 'some.answer', 42)
        class Foo(object):
            def call(self):
                print(self._answer)  # Prints 42

                # The properties of the services provided by this component
                # instance would be updated during this assignation
                self._answer = 100
    """
    HANDLER_ID = constants.HANDLER_PROPERTY
    """ ID of the handler configured by this decorator """

    def __init__(self, field, name=None, value=None):
        """
        :param field: The property field in the class (can't be ``None`` nor
                      empty)
        :param name: The property name (if ``None``, this will be the field
                     name)
        :param value: The property value (``None`` by default)
        :raise TypeError: Invalid argument type
        :raise ValueError: If the name or the name is ``None`` or empty
        """
        # Field validity test
        if not is_string(field):
            raise TypeError("Field name must be a string")

        field = field.strip()
        if not field or " " in field:
            raise ValueError(
                "Empty or invalid property field name '{0}'".format(field)
            )

        # Name validity test
        if name is not None:
            if not is_string(name):
                raise TypeError("Property name must be a string")

            name = name.strip()

        if not name:
            # No name given: use the field name
            name = field

        self._field = field
        self._name = name
        self._value = value

    def __call__(self, clazz):
        """
        Adds the property to the class iPOPO properties field.
        Creates the field if needed.

        :param clazz: The class to decorate
        :return: The decorated class
        :raise TypeError: If *clazz* is not a type
        """
        if not inspect.isclass(clazz):
            raise TypeError(
                "@Property can decorate only classes, not '{0}'".format(
                    type(clazz).__name__
                )
            )

        # Get the factory context
        context = get_factory_context(clazz)
        if context.completed:
            # Do nothing if the class has already been manipulated
            _logger.warning(
                "@Property: Already manipulated class: %s",
                get_method_description(clazz),
            )
            return clazz

        # Set up the property in the class
        context.properties[self._name] = self._value

        # Associate the field to the property name
        context.properties_fields[self._field] = self._name

        # Mark the handler in the factory context
        context.set_handler(self.HANDLER_ID, None)

        # Inject a property in the class. The property will call an instance
        # level getter / setter, injected by iPOPO after the instance creation
        setattr(
            clazz,
            self._field,
            _ipopo_class_field_property(
                self._name, self._value, constants.IPOPO_PROPERTY_PREFIX
            ),
        )

        return clazz


class HiddenProperty(Property):
    # pylint: disable=R0903
    """
    The ``@HiddenProperty`` decorator defines a component property which won't
    be visible in the properties of the services it provides.
    This kind of property is also not accessible using iPOPO reflection
    methods.

    This decorator has the same handler, accepts the same parameters and
    follows the same rules as the :class:`Property` decorator.

    :Handler ID: :py:const:`pelix.ipopo.constants.HANDLER_PROPERTY`

    :Example:

    .. code-block:: python

        @ComponentFactory()
        @HiddenProperty('_password', 'some.password', "secret")
        class Foo(object):
            def call(self):
                print(self._password)  # I think we're missing the point

                # Service properties won't be affected by this change
                self._password = "UpdatedSecret"
    """

    def __call__(self, clazz):
        """
        Adds the property to the class iPOPO properties field.
        Creates the field if needed.

        :param clazz: The class to decorate
        :return: The decorated class
        :raise TypeError: If *clazz* is not a type
        """
        if not inspect.isclass(clazz):
            raise TypeError(
                "@HiddenProperty can decorate only classes, not '{0}'".format(
                    type(clazz).__name__
                )
            )

        # Get the factory context
        context = get_factory_context(clazz)
        if context.completed:
            # Do nothing if the class has already been manipulated
            _logger.warning(
                "@HiddenProperty: Already manipulated class: %s",
                get_method_description(clazz),
            )
            return clazz

        # Set up the property in the class
        context.hidden_properties[self._name] = self._value

        # Mark the handler in the factory context
        context.set_handler(self.HANDLER_ID, None)

        # Inject a property in the class. The property will call an instance
        # level getter / setter, injected by iPOPO after the instance creation
        setattr(
            clazz,
            self._field,
            _ipopo_class_field_property(
                self._name, self._value, constants.IPOPO_HIDDEN_PROPERTY_PREFIX
            ),
        )

        return clazz


# ------------------------------------------------------------------------------


def _get_specifications(specifications):
    """
    Computes the list of strings corresponding to the given specifications

    :param specifications: A string, a class or a list of specifications
    :return: A list of strings
    :raise ValueError: Invalid specification found
    """
    if not specifications or specifications is object:
        raise ValueError("No specifications given")
    elif inspect.isclass(specifications):
        if Provides.USE_MODULE_QUALNAME:
            if sys.version_info < (3, 3, 0):
                raise ValueError(
                    "Qualified name capability requires Python 3.3+"
                )

            # Get the name of the class
            if not specifications.__module__:
                return [specifications.__qualname__]

            return [
                "{0}.{1}".format(
                    specifications.__module__, specifications.__qualname__
                )
            ]
        else:
            # Legacy behavior
            return [specifications.__name__]
    elif is_string(specifications):
        # Specification name
        specifications = specifications.strip()
        if not specifications:
            raise ValueError("Empty specification given")
        return [specifications]
    elif isinstance(specifications, (list, tuple)):
        # List given: normalize its content
        results = []
        for specification in specifications:
            results.extend(_get_specifications(specification))
        return results
    else:
        raise ValueError(
            "Unhandled specifications type : {0}".format(
                type(specifications).__name__
            )
        )


class Provides(object):
    # pylint: disable=R0903
    """
    The ``@Provides`` decorator defines a service to be exposed by component
    instances.
    This service will be registered (visible) in the Pelix service registry
    while the component is valid and its service controller is set to ``True``.

    All the properties of the component defined with the :class:`Property`
    decorator will be visible in the service properties.

    The controller is an injected field (a Python *property*) that must contain
    a boolean.
    By default, the controller is set to ``True``, *i.e.* the service will be
    provided by the component when it is validated.

    :Handler ID: :py:const:`pelix.ipopo.constants.HANDLER_PROVIDES`

    :Example:

    .. code-block:: python

        @ComponentFactory()
        # "answer.prefix" will be a property of the service
        @Property("_answer", "answer.prefix", "Hello")
        @Provides("hello.world")
        class Foo(object):
            # The component instance will publish a "hello.world" service as
            # long as it is valid
            def greet(self, name):
                print(self._answer, name, "!")

        @ComponentFactory()
        # This service will provide multiple specifications
        @Provides(["hello.world", "hello.world.extended"], "_svc_flag")
        @Provides("reset")
        class Bar(object):
            def greet(self, name):
                # Implementation of hello.world
                print("Hello,", name, "!")

            def adieu(self, name):
                print("So long,", name, "!")

                # Sets the controller to False: the service won't be published
                # anymore until the controller is set back to True
                self._svc_flag = False

            def reset(self):
                # Implementation of the "reset" service: publish the service
                # again
                self._svc_flag = True
    """
    HANDLER_ID = constants.HANDLER_PROVIDES
    """ ID of the handler configured by this decorator """

    USE_MODULE_QUALNAME = False
    """
    Selects the methodology to generate a specification from a class.
    A value of False uses __name__ (legacy), while True enables
    __name__ + '.' + __qualname__
    """

    def __init__(
        self, specifications, controller=None, factory=False, prototype=False
    ):
        """
        :param specifications: A list of provided specification(s), or the
                               single provided specification (can't be empty)
        :param controller: The name of the service controller class field
                           (optional)
        :param factory: If True, this service is a service factory
                        (False by default)
        :param prototype: If True, this service is prototype service factory
                          (False by default)
        :raise ValueError: If the specifications are invalid
        """
        if controller is not None:
            if not is_string(controller):
                raise ValueError("Controller name must be a string")

            controller = controller.strip()
            if not controller:
                # Empty controller name
                _logger.warning("Empty controller name given")
                controller = None
            elif " " in controller:
                raise ValueError("Controller name contains spaces")

        self.__specifications = specifications
        self.__controller = controller
        self.__is_factory = factory
        self.__is_prototype = prototype

    def __call__(self, clazz):
        """
        Adds the provided service information to the class context iPOPO field.
        Creates the field if needed.

        :param clazz: The class to decorate
        :return: The decorated class
        :raise TypeError: If *clazz* is not a type or if the service factory
                          methods are missing
        """
        if not inspect.isclass(clazz):
            raise TypeError(
                "@Provides can decorate only classes, not '{0}'".format(
                    type(clazz).__name__
                )
            )

        # Get the factory context
        context = get_factory_context(clazz)
        if context.completed:
            # Do nothing if the class has already been manipulated
            _logger.warning(
                "@Provides: Already manipulated class: %s",
                get_method_description(clazz),
            )
            return clazz

        # Avoid duplicates (but keep the order)
        filtered_specs = []
        if not self.__specifications:
            filtered_specs = _get_specifications(clazz.__bases__)
        else:
            # Avoid duplicates (but keep the order)
            specs = _get_specifications(self.__specifications)
            for spec in specs:
                if spec not in filtered_specs:
                    filtered_specs.append(spec)

        # Store the service information
        config = context.set_handler_default(self.HANDLER_ID, [])
        config.append(
            (
                filtered_specs,
                self.__controller,
                self.__is_factory,
                self.__is_prototype,
            )
        )

        if self.__controller:
            # Inject a property in the class. The property will call an
            # instance level getter / setter, injected by iPOPO after the
            # instance creation
            setattr(
                clazz,
                self.__controller,
                _ipopo_class_field_property(
                    self.__controller, True, constants.IPOPO_CONTROLLER_PREFIX
                ),
            )

            # Inject the future controller methods
            setattr(
                clazz,
                constants.IPOPO_CONTROLLER_PREFIX
                + constants.IPOPO_GETTER_SUFFIX,
                None,
            )
            setattr(
                clazz,
                constants.IPOPO_CONTROLLER_PREFIX
                + constants.IPOPO_SETTER_SUFFIX,
                None,
            )

        if self.__is_factory or self.__is_prototype:
            # Ensure that the service factory methods exist
            try:
                validate_method_arity(
                    clazz.get_service, "bundle", "service_registration"
                )
                validate_method_arity(
                    clazz.unget_service, "bundle", "service_registration"
                )
            except AttributeError as ex:
                raise TypeError(
                    "Service factories must provide an {} method".format(ex)
                )

        if self.__is_prototype:
            # Ensure that the prototype service factory methods exist
            try:
                validate_method_arity(
                    clazz.unget_service_instance,
                    "bundle",
                    "service_registration",
                    "service",
                )
            except AttributeError as ex:
                raise TypeError(
                    "Prototype Service factories must provide "
                    "an {} method".format(ex)
                )

        return clazz


# ------------------------------------------------------------------------------


class Requires(object):
    # pylint: disable=R0903
    """
    The ``@Requires`` decorator defines the requirement of a service.

    :Handler ID: :py:const:`pelix.ipopo.constants.HANDLER_REQUIRES`

    :Example:

    .. code-block:: python

        @ComponentFactory()
        @Requires('_hello', 'hello.world')
        class Foo(object):
            pass

        @ComponentFactory()
        @Requires('_hello', 'hello.world', aggregate=True, optional=False,
                  spec_filter='(language=fr)')
        class Bar(object):
            pass
    """
    HANDLER_ID = constants.HANDLER_REQUIRES
    """ ID of the handler configured by this decorator """

    def __init__(
        self,
        field,
        specification,
        aggregate=False,
        optional=False,
        spec_filter=None,
        immediate_rebind=False,
    ):
        """
        :param field: The field where to inject the requirement
        :param specification: The specification of the service to inject
        :param aggregate: If True, injects a list of services, else the first
                          matching service
        :param optional: If True, this injection is optional: the component can
                         be valid without it
        :param spec_filter: An LDAP query to filter injected services according
                            to their properties
        :param immediate_rebind:
            If True, the component won't be invalidated then re-validated if a
            matching service is available when the injected dependency is
            unbound

        The ``field`` and ``specification`` parameters are mandatory.
        By default, a requirement is neither aggregated nor optional
        (both are set to ``False``) and no specification filter is used.

        .. note:: Since iPOPO 0.5.4, only one specification can be given.
        """
        if not field:
            raise ValueError("Empty field name.")

        if not is_string(field):
            raise TypeError(
                "The field name must be a string, not {0}".format(
                    type(field).__name__
                )
            )

        if " " in field:
            raise ValueError("Field name can't contain spaces.")

        self._field = field

        # Be sure that there is only one required specification
        specifications = _get_specifications(specification)
        self._multi_specs = len(specifications) > 1

        # Construct the requirement object
        self._requirement = Requirement(
            specifications[0],
            aggregate,
            optional,
            spec_filter,
            immediate_rebind,
        )

    def __call__(self, clazz):
        """
        Adds the requirement to the class iPOPO field

        :param clazz: The class to decorate
        :return: The decorated class
        :raise TypeError: If *clazz* is not a type
        """
        if not inspect.isclass(clazz):
            raise TypeError(
                "@{0} can decorate only classes, not '{1}'".format(
                    type(self).__name__, type(clazz).__name__
                )
            )

        if self._multi_specs:
            _logger.warning(
                "%s: Only one specification can be required: %s -> %s",
                type(self).__name__,
                clazz.__name__,
                self._field,
            )

        # Set up the property in the class
        context = get_factory_context(clazz)
        if context.completed:
            # Do nothing if the class has already been manipulated
            _logger.warning(
                "@%s: Already manipulated class: %s",
                type(self).__name__,
                get_method_description(clazz),
            )
            return clazz

        # Store the requirement information
        config = context.set_handler_default(self.HANDLER_ID, {})
        config[self._field] = self._requirement

        # Inject the field
        setattr(clazz, self._field, None)
        return clazz


# ------------------------------------------------------------------------------


class RequiresVarFilter(Requires):
    # pylint: disable=R0903
    """
    The ``@RequiresVarFilter`` decorator acts like :class:`Requires`, but its
    LDAP filter dynamically adapts to the properties of this component.

    The decorator accepts a component property key using the ``{key}`` format
    in the LDAP filter string. This placeholder will be replaced by the
    property value, and will the filter will be updated each time the property
    value changes.

    :Handler ID:
            :py:const:`pelix.ipopo.constants.HANDLER_REQUIRES_VARIABLE_FILTER`

    :Example:

    .. code-block:: python

        @ComponentFactory()
        @Property("_lang", "lang", "fr")
        @RequiresVarFilter("_hello", "hello.world", optional=True,
                           spec_filter="(language={lang})")
        class Bar(object):
            def call(self):
                # The dependency is optional
                if self._hello is not None:
                    # Default "lang" instance property is set to "fr"
                    self._hello.greet("le Monde")  # Bonjour le Monde

                # Change the property to have another service
                self._lang = "en"

                # We can call the new service immediately as the dependency is
                # optional (no risk of invalidation), but we have to check if
                # such a service exists
                if self._hello is not None:
                    self._hello.greet("World")  # Hello World
    """
    HANDLER_ID = constants.HANDLER_REQUIRES_VARIABLE_FILTER
    """ ID of the handler configured by this decorator """


# ------------------------------------------------------------------------------


class RequiresBest(Requires):
    # pylint: disable=R0903
    """
    The ``@RequiresBest`` decorator acts like :class:`Requires`, but it
    always injects the service with the best rank (``service.ranking``
    property).

    Unlike most of the other requirement decorators, ``@RequiresBest`` doesn't
    support the injection of a list of services (``aggregate``) as only the
    best service is injected.

    :Handler ID: :py:const:`pelix.ipopo.constants.HANDLER_REQUIRES_BEST`

    :Example:

    .. code-block:: python

        @ComponentFactory()
        @RequiresBest('_hello', 'hello.world', immediate_rebind=True)
        class Foo(object):
            def call(self):
                # First call, with the current best service
                self._hello.greet("World")  # prints "Hello, World!"

                # Something happens, the rank of "hello.world" services changes
                # For example, locale changed and French now has a higher rank
                time.sleep(1)

                # Second call without waiting for re-validation as we have set
                # immediate_rebind=True for this example
                self._hello.greet("World")  # prints "Bonjour, World !"

        # We can also use a specification filter or make the service optional
        @ComponentFactory()
        @RequiresBest('_hello', 'hello.world', optional=True,
                      spec_filter='(language=fr)')
        class Bar(object):
            def call(self):
                if self._hello is not None:
                    # First call, with the current best service
                    self._hello.greet("World")  # prints "Hello, World!"
    """
    HANDLER_ID = constants.HANDLER_REQUIRES_BEST
    """ ID of the handler configured by this decorator """

    def __init__(
        self,
        field,
        specification,
        optional=False,
        spec_filter=None,
        immediate_rebind=True,
    ):
        """
        :param field: The injected field
        :param specification: The injected service specification
        :param optional: If True, this injection is optional
        :param spec_filter: An LDAP query to filter injected services upon
                            their properties
        :param immediate_rebind: If True, the component won't be invalidated
                                 then re-validated if a matching service is
                                 available when the injected dependency is
                                 unbound
        :raise TypeError: A parameter has an invalid type
        :raise ValueError: An error occurred while parsing the filter or an
                           argument is incorrect
        """
        super(RequiresBest, self).__init__(
            field, specification, False, optional, spec_filter, immediate_rebind
        )


# ------------------------------------------------------------------------------


class RequiresMap(Requires):
    # pylint: disable=R0903
    """
    The ``@RequiresMap`` decorator defines a requirement that must be injected
    in a dictionary, based on a service property.

    :Handler ID: :py:const:`pelix.ipopo.constants.HANDLER_REQUIRES_MAP`

    :Example:

    .. code-block:: python

        @ComponentFactory()
        @RequiresMap("_hello", "hello.world", "language")
        class Bar(object):
            def call(self):
                self._hello["en"].hello("World")
                self._hello["fr"].hello("le Monde")
    """
    HANDLER_ID = constants.HANDLER_REQUIRES_MAP
    """ ID of the handler configured by this decorator """

    def __init__(
        self,
        field,
        specification,
        key,
        allow_none=False,
        aggregate=False,
        optional=False,
        spec_filter=None,
    ):
        """
        :param field: The injected field
        :param specification: The injected service specification
        :param key: The name of the service property to use as a dictionary key
        :param allow_none: If True, also injects services with the property
                           value set to ``None`` or missing
        :param aggregate: If True, injects a list of services with the same
                          property value, else injects only one service
                          (first found) per property value
        :param optional: If True, this injection is optional
        :param spec_filter: An LDAP query to filter injected services upon
                            their properties
        :raise TypeError: A parameter has an invalid type
        :raise ValueError: An error occurred while parsing the filter or an
                           argument is incorrect
        """
        super(RequiresMap, self).__init__(
            field, specification, aggregate, optional, spec_filter, False
        )
        # Check if key is valid
        if not key:
            raise ValueError("No property key given")

        # Store the flags
        self._key = key
        self._allow_none = allow_none

    def __call__(self, clazz):
        """
        Adds the requirement to the class iPOPO field

        :param clazz: The class to decorate
        :return: The decorated class
        :raise TypeError: If *clazz* is not a type
        """
        clazz = super(RequiresMap, self).__call__(clazz)

        # Set up the property in the class
        context = get_factory_context(clazz)
        if not context.completed:
            # Store the requirement information
            config = context.set_handler_default(self.HANDLER_ID, {})
            config[self._field] = (
                self._requirement,
                self._key,
                self._allow_none,
            )
        return clazz


# ------------------------------------------------------------------------------


class RequiresBroadcast(Requires):
    # pylint: disable=R0903
    """
    The ``@RequiresBroadcast`` decorator defines a requirement that will be
    injected as a single object, hiding the underlying missing dependency or
    group of services matching the requirement.

    Unlike :class:`Requires`, the parameter ``optional`` is set to ``True`` by
    default. Also, the ``aggregate`` argument is not available, the behaviour
    of this handler is to broadcast to all matching services.

    :Handler ID: :py:const:`pelix.ipopo.constants.HANDLER_REQUIRES_BRODCAST`

    :Example:

    .. code-block:: python

        @ComponentFactory()
        @RequiresBroadcast("_notifier", "some.notifier")
        class Bar(object):
            def trace(self, message):
                # We can use the service as a single object, without taking
                # care of the number of services matching our requirement:
                self._notifier.notify("Hello, world")
    """
    HANDLER_ID = constants.HANDLER_REQUIRES_BRODCAST
    """ ID of the handler configured by this decorator """

    def __init__(
        self,
        field,
        specification,
        optional=True,
        spec_filter=None,
        muffle_exceptions=True,
        trace_exceptions=True,
    ):
        """
        :param field: The injected field
        :param specification: The injected service specification
        :param optional: If True, this injection is optional
        :param spec_filter: An LDAP query to filter injected services upon
                            their properties
        :param muffle_exceptions: If True, exceptions raised by underlying
                                  services are not propagated (True by default)
        :param trace_exceptions: If True, trace the exceptions that are muffled
                                 (True by default)
        :raise TypeError: A parameter has an invalid type
        :raise ValueError: An error occurred while parsing the filter or an
                           argument is incorrect
        """
        # Forced flags: aggregate and immediate rebind
        super(RequiresBroadcast, self).__init__(
            field, specification, True, optional, spec_filter, True
        )

        # Store the flags
        self._muffle_ex = muffle_exceptions
        self._trace_ex = trace_exceptions

    def __call__(self, clazz):
        """
        Adds the requirement to the class iPOPO field

        :param clazz: The class to decorate
        :return: The decorated class
        :raise TypeError: If *clazz* is not a type
        """
        clazz = super(RequiresBroadcast, self).__call__(clazz)

        # Set up the property in the class
        context = get_factory_context(clazz)
        if not context.completed:
            # Store the requirement information
            config = context.set_handler_default(self.HANDLER_ID, {})
            config[self._field] = (
                self._requirement,
                self._muffle_ex,
                self._trace_ex,
            )
        return clazz


# ------------------------------------------------------------------------------


class Temporal(Requires):
    # pylint: disable=R0903
    """
    The ``@Temporal`` decorator defines a single immediate rebind requirement
    with a grace time when the injected service disappears.

    This decorator acts like :class:`Requires` except it doesn't support the
    ``immediate_rebind`` (forced to ``True``) nor ``aggregate`` arguments.

    When the injected service disappears, the component won't be invalidated
    before the given timeout.
    If a matching is found, it is injected in-place and the component instance
    continues its operations.
    If the service is used while no service is available, the call is put in
    hold and blocks until a new service is injected or until the timeout is
    reached. In the latter case, a ``TemporalException`` is raised.

    :Handler ID: :py:const:`pelix.ipopo.constants.HANDLER_TEMPORAL`

    :Example:

    .. code-block:: python

        @ComponentFactory()
        @Temporal('_hello', 'hello.world', timeout=5)
        class Bar(object):
            def call(self):
                # If the service is injected: this call is immediate
                # If the service has gone away, this call will hold for
                # 5 seconds (timeout parameter)
                # - if a service is injected during the grace period, the call
                #   will be done
                try:
                    self._hello.greet("World")
                except pelix.ipopo.handlers.temporal.TemporalException:
                    # - else, a TemporalException is raised
                    print("Service disappeared")
    """
    HANDLER_ID = constants.HANDLER_TEMPORAL
    """ ID of the handler configured by this decorator """

    def __init__(
        self, field, specification, optional=False, spec_filter=None, timeout=10
    ):
        """
        :param field: The injected field
        :param specification: The injected service specification
        :param optional: If True, this injection is optional
        :param spec_filter: An LDAP query to filter injected services upon
                            their properties
        :param timeout: Temporal timeout, in seconds (must be greater than 0)
        :raise TypeError: A parameter has an invalid type
        :raise ValueError: An error occurred while parsing the filter or an
                           argument is incorrect
        """
        super(Temporal, self).__init__(
            field, specification, False, optional, spec_filter, True
        )
        if timeout <= 0:
            _logger.warning(
                "@Temporal timeout must be greater than 0. "
                "Using default value."
            )
            self._timeout = 10
        else:
            self._timeout = timeout

    def __call__(self, clazz):
        """
        Adds the requirement to the class iPOPO field

        :param clazz: The class to decorate
        :return: The decorated class
        :raise TypeError: If *clazz* is not a type
        """
        clazz = super(Temporal, self).__call__(clazz)

        # Store the requirement information
        context = get_factory_context(clazz)
        if not context.completed:
            config = context.set_handler_default(self.HANDLER_ID, {})
            config[self._field] = (self._requirement, self._timeout)
        return clazz


# ------------------------------------------------------------------------------


class BindField(object):
    # pylint: disable=R0903
    """
    The ``@BindField`` callback decorator is called when a component is bound
    to a dependency, injected in the given field.

    The decorated method must accept the field where the service has been
    injected, the service object and its
    :class:`~pelix.framework.ServiceReference` as arguments.

    If the service is a required one, the bind callback is called **before**
    the component is validated.
    The bind field callback is called **after** the global bind method.

    The service reference can be stored *if it is released on unbind*.

    Exceptions raised by a bind callback are ignored.

    :Example:

    .. code-block:: python

       @ComponentFactory()
       @Requires("_hello", "hello.svc")
       class Foo:
          @BindField("_hello")
          def bind_method(self, field, service, service_reference):
              '''
              field: Field wherein the dependency is injected ("_hello")
              service: The injected service instance
              service_reference: The injected service ServiceReference
              '''
              # ...
    """

    def __init__(self, field, if_valid=False):
        """
        :param field: The field associated to the binding
        :param if_valid: If True, call the decorated method only when the
                         component is valid
        """
        self._field = field
        self._if_valid = if_valid

    def __call__(self, method):
        """
        Updates the "field callback" list for this method

        :param method: Method to decorate
        :return: Decorated method
        :raise TypeError: The decorated element is not a valid function
        """
        if not inspect.isroutine(method):
            raise TypeError("@BindField can only be applied on functions")

        # Tests the number of parameters
        validate_method_arity(method, "field", "service", "service_reference")

        _append_object_entry(
            method,
            constants.IPOPO_METHOD_FIELD_CALLBACKS,
            (constants.IPOPO_CALLBACK_BIND_FIELD, self._field, self._if_valid),
        )
        return method


class UpdateField(object):
    # pylint: disable=R0903
    """
    The ``@UpdateField`` callback decorator is called when the properties of
    a service injected in the given field have been updated.

    The decorated method must accept the field where the service has been
    injected, the service object, its
    :class:`~pelix.framework.ServiceReference` and its previous properties as
    arguments.

    Exceptions raised by an update callback are ignored.

    :Example:

    .. code-block:: python

       @ComponentFactory()
       @Requires("_hello", "hello.svc")
       class Foo:
          @UpdateField("_hello")
          def update_method(self, service, service_reference, old_properties):
              '''
              field: Field wherein the dependency was injected ("_hello")
              service: The injected service instance
              service_reference: The injected service ServiceReference
              old_properties: The previous service properties
              '''
              # ...
    """

    def __init__(self, field, if_valid=False):
        """
        :param field: The field associated to the binding
        :param if_valid: If True, call the decorated method only when the
                         component is valid
        """
        self._field = field
        self._if_valid = if_valid

    def __call__(self, method):
        """
        Updates the "field callback" list for this method

        :param method: Method to decorate
        :return: Decorated method
        :raise TypeError: The decorated element is not a valid function
        """
        if not inspect.isroutine(method):
            raise TypeError("@UnbindField can only be applied on functions")

        # Tests the number of parameters
        validate_method_arity(
            method, "field", "service", "service_reference", "old_properties"
        )

        _append_object_entry(
            method,
            constants.IPOPO_METHOD_FIELD_CALLBACKS,
            (
                constants.IPOPO_CALLBACK_UPDATE_FIELD,
                self._field,
                self._if_valid,
            ),
        )
        return method


class UnbindField(object):
    # pylint: disable=R0903
    """
    The ``@UnbindField`` callback decorator is called when an injected
    dependency is unbound.

    The decorated method must accept the field where the service has been
    injected, the service object, its
    :class:`~pelix.framework.ServiceReference` and its previous properties as
    arguments.

    If the service is a required one, the unbind callback is called **after**
    the component has been invalidated.
    The unbind field callback is called **before** the global unbind method.

    Exceptions raised by an unbind callback are ignored.

    :Example:

    .. code-block:: python

       @ComponentFactory()
       @Requires("_hello", "hello.svc")
       class Foo:
          @UnbindField("_hello")
          def unbind_method(self, field, service, service_reference):
              '''
              field: Field wherein the dependency was injected ("_hello")
              service: The injected service instance
              service_reference: The injected service ServiceReference
              '''
              # ...
    """

    def __init__(self, field, if_valid=False):
        """
        :param field: The field associated to the binding
        :param if_valid: If True, call the decorated method only when the
                         component is valid
        """
        self._field = field
        self._if_valid = if_valid

    def __call__(self, method):
        """
        Updates the "field callback" list for this method

        :param method: Method to decorate
        :return: Decorated method
        :raise TypeError: The decorated element is not a valid function
        """
        if not inspect.isroutine(method):
            raise TypeError("@UnbindField can only be applied on functions")

        # Tests the number of parameters
        validate_method_arity(method, "field", "service", "service_reference")

        _append_object_entry(
            method,
            constants.IPOPO_METHOD_FIELD_CALLBACKS,
            (
                constants.IPOPO_CALLBACK_UNBIND_FIELD,
                self._field,
                self._if_valid,
            ),
        )
        return method


# ------------------------------------------------------------------------------


def Bind(method):
    # pylint: disable=C0103
    """
    The ``@Bind`` callback decorator is called when a component is bound to a
    dependency.

    The decorated method must accept the injected service object and its
    :class:`~pelix.framework.ServiceReference` as arguments.

    If the service is a required one, the bind callback is called **before**
    the component is validated.

    The service reference can be stored *if it is released on unbind*.

    Exceptions raised by a bind callback are ignored.

    :Example:

    .. code-block:: python

       @ComponentFactory()
       @Requires("_hello", "hello.svc")
       class Foo:
           @Bind
           def bind_method(self, service, service_reference):
               '''
               service: The injected service instance.
               service_reference: The injected service ServiceReference
               '''
               # ...

    :param method: The decorated method
    :raise TypeError: The decorated element is not a valid function
    """
    if not inspect.isroutine(method):
        raise TypeError("@Bind can only be applied on functions")

    # Tests the number of parameters
    validate_method_arity(method, "service", "service_reference")

    _append_object_entry(
        method, constants.IPOPO_METHOD_CALLBACKS, constants.IPOPO_CALLBACK_BIND
    )
    return method


def Update(method):
    # pylint: disable=C0103
    """
    The ``@Update`` callback decorator is called when the properties of an
    injected service have been modified.

    The decorated method must accept the injected service object and its
    :class:`~pelix.framework.ServiceReference` and the previous properties
    as arguments.

    Exceptions raised by an update callback are ignored.

    :Example:

    .. code-block:: python

       @ComponentFactory()
       @Requires("_hello", "hello.svc")
       class Foo:
           @Update
           def update_method(self, service, service_reference, old_properties):
               '''
               service: The injected service instance.
               service_reference: The injected service ServiceReference
               old_properties: The previous service properties
               '''
               # ...

    :param method: The decorated method
    :raise TypeError: The decorated element is not a valid function
    """
    if not isinstance(method, types.FunctionType):
        raise TypeError("@Update can only be applied on functions")

    # Tests the number of parameters
    validate_method_arity(
        method, "service", "service_reference", "old_properties"
    )

    _append_object_entry(
        method,
        constants.IPOPO_METHOD_CALLBACKS,
        constants.IPOPO_CALLBACK_UPDATE,
    )
    return method


def Unbind(method):
    # pylint: disable=C0103
    """
    The ``@Unbind`` callback decorator is called when a component dependency is
    unbound.

    The decorated method must accept the injected service object and its
    :class:`~pelix.framework.ServiceReference` as arguments.

    If the service is a required one, the unbind callback is called **after**
    the component has been invalidated.

    Exceptions raised by an unbind callback are ignored.

    :Example:

    .. code-block:: python

       @ComponentFactory()
       @Requires("_hello", "hello.svc")
       class Foo:
           @Unbind
           def unbind_method(self, service, service_reference):
               '''
               service: The previously injected service instance.
               service_reference: Its ServiceReference
               '''
               # ...

    :param method: The decorated method
    :raise TypeError: The decorated element is not a valid function
    """
    if not isinstance(method, types.FunctionType):
        raise TypeError("@Unbind can only be applied on functions")

    # Tests the number of parameters
    validate_method_arity(method, "service", "service_reference")

    _append_object_entry(
        method,
        constants.IPOPO_METHOD_CALLBACKS,
        constants.IPOPO_CALLBACK_UNBIND,
    )
    return method


# ------------------------------------------------------------------------------


class ValidateComponent(object):
    # pylint: disable=R0903
    """
    The ``@ValidateComponent`` decorator declares a callback method for
    component validation.

    Currently, the arguments given to the callback are read-only, to avoid
    messing with the validation life-cycle.
    In the future, it might be possible to modify the properties and to use
    the component context in order to customize the component early.

    :Example:

    Here are some sample uses of the decorator. Note that the number and order
    of arguments only has to match the list given to the decorator:

    .. code-block:: python

       from pelix.constants import ARG_COMPONENT_CONTEXT, \\
           ARG_BUNDLE_CONTEXT, ARG_PROPERTIES

       @ValidateComponent(ARG_COMPONENT_CONTEXT)
       def validate_component(self, component_ctx):
           # ...

       @ValidateComponent(ARG_BUNDLE_CONTEXT, ARG_COMPONENT_CONTEXT)
       def validate_component(self, bundle_ctx, component_ctx):
           # ...

       @ValidateComponent(
           ARG_BUNDLE_CONTEXT, ARG_COMPONENT_CONTEXT, ARG_PROPERTIES
        )
       def validate_component(self, bundle_ctx, component_ctx, props):
           # ...
    """

    def __init__(self, *args):
        """
        :param args: The decorator accepts an ordered list of arguments.
                     They define the signature of the decorated method.

                     The arguments can be the following ones, declared in the
                     ``pelix.ipopo.constants`` module:

                     * ``ARG_BUNDLE_CONTEXT``: Gives access to the bundle
                       context
                     * ``ARG_COMPONENT_CONTEXT``: Gives access to the
                       component context
                     * ``ARG_PROPERTIES``: Gives access to the initial
                       properties of the component (``dict``)

        :raise TypeError: A parameter has an invalid type or the decorated
                          object is not a method
        """
        # Check arguments validity
        valid_args = (
            constants.ARG_BUNDLE_CONTEXT,
            constants.ARG_COMPONENT_CONTEXT,
            constants.ARG_PROPERTIES,
        )

        for arg in args:
            if arg not in valid_args:
                raise TypeError("Unknown argument type: {}".format(arg))

        # Keep track of the arguments
        self._args = tuple(args)

    def __call__(self, method):
        """
        Registers the decorated method as a callback for component validation

        :param method: The validation method
        :raise TypeError: The decorated element is not a valid function
        """
        if not isinstance(method, types.FunctionType):
            raise TypeError(
                "@ValidateComponent can only be applied on functions"
            )

        # Tests the number of parameters
        validate_method_arity(method, *self._args)

        # Append the callback to the component
        _append_object_entry(
            method,
            constants.IPOPO_METHOD_CALLBACKS,
            constants.IPOPO_CALLBACK_VALIDATE,
        )

        # Append arguments list to the method
        _set_object_entry(method, constants.IPOPO_VALIDATE_ARGS, self._args)

        return method


class InvalidateComponent(ValidateComponent):
    # pylint: disable=R0903
    """
    The ``@InvalidateComponent`` decorator declares a callback method for
    component invalidation.

    Its arguments and their order describes the ones of the callback it
    decorates.
    They are the same as those of :class:`ValidateComponent`.

    Exceptions raised by an invalidation callback are ignored.

    If the component provides a service, the invalidation method is called
    after the provided service has been unregistered to the framework.
    """

    def __call__(self, method):
        """
        Registers the decorated method as a callback for component invalidation

        :param method: The invalidation method
        :raise TypeError: The decorated element is not a valid function
        """
        if not isinstance(method, types.FunctionType):
            raise TypeError(
                "@InvalidateComponent can only be applied on functions"
            )

        # Tests the number of parameters
        validate_method_arity(method, *self._args)

        # Append the callback to the component
        _append_object_entry(
            method,
            constants.IPOPO_METHOD_CALLBACKS,
            constants.IPOPO_CALLBACK_INVALIDATE,
        )

        # Append arguments list to the method
        _set_object_entry(method, constants.IPOPO_VALIDATE_ARGS, self._args)

        return method


# ------------------------------------------------------------------------------


def Validate(method):
    # pylint: disable=C0103
    """
    This decorator is an alias to :class:`ValidateComponent` to decorate a
    callback method than only accepts the
    :class:`~pelix.framework.BundleContext` argument.
    It is not possible to have both ``@Validate`` and ``@ValidateComponent``
    decorators used in the same class.

    The validation callback decorator is called when a component becomes valid,
    *i.e.* if all of its required dependencies has been injected.

    If the validation callback raises an exception, the component goes into
    **ERRONEOUS** state.

    If the component provides a service, the validation method is called before
    the provided service is registered to the framework.

    :Example:

    .. code-block:: python

       @ComponentFactory()
       class Foo:
           @Validate
           def validation_method(self, bundle_context):
               '''
               bundle_context: The component's bundle context
               '''
               # ...

    :param method: The validation callback method
    :raise TypeError: The decorated element is not a valid function
    """
    return ValidateComponent(constants.ARG_BUNDLE_CONTEXT)(method)


def Invalidate(method):
    # pylint: disable=C0103
    """
    This decorator is an alias to :class:`InvalidateComponent` to decorate a
    callback method than only accepts the
    :class:`~pelix.framework.BundleContext` argument.
    It is not possible to have both ``@Invalidate`` and
    ``@InvalidateComponent`` decorators used in the same class.

    The invalidation callback decorator is called when a component becomes
    invalid, *i.e.* if one of its required dependencies disappeared.

    Exceptions raised by an invalidation callback are ignored.

    If the component provides a service, the invalidation method is called
    after the provided service has been unregistered to the framework.

    :Example:

    .. code-block:: python

       @ComponentFactory()
       class Foo:
           @Invalidate
           def invalidation_method(self, bundle_context):
               '''
               bundle_context: The component's bundle context
               '''
               # ...

    :param method: The decorated method
    :raise TypeError: The decorated element is not a function
    """
    return InvalidateComponent(constants.ARG_BUNDLE_CONTEXT)(method)


def PostRegistration(method):
    # pylint: disable=C0103
    """
    The service post-registration callback decorator is called after a service
    of the component has been registered to the framework.

    The decorated method must accept the
    :class:`~pelix.framework.ServiceReference` of the registered
    service as argument.

    Note that when this method is called, the consumers have already been bound
    and might already use the service.

    :Example:

    .. code-block:: python

       @ComponentFactory()
       @Provides("hello.svc")
       class Foo:
           @PostRegistration
           def callback_method(self, service_reference):
               '''
               service_reference: The ServiceReference of the provided service
               '''
               # Custom notification, ...

    :param method: The decorated method
    :raise TypeError: The decorated element is not a valid function
    """
    if not isinstance(method, types.FunctionType):
        raise TypeError("@PostRegistration can only be applied on functions")

    # Tests the number of parameters
    validate_method_arity(method, "service_reference")
    _append_object_entry(
        method,
        constants.IPOPO_METHOD_CALLBACKS,
        constants.IPOPO_CALLBACK_POST_REGISTRATION,
    )
    return method


def PostUnregistration(method):
    # pylint: disable=C0103
    """
    The service post-unregistration callback decorator is called after a
    service of the component has been unregistered from the framework.

    The decorated method must accept the
    :class:`~pelix.framework.ServiceReference` of the registered
    service as argument.

    This method being called after the unregistration, the consumers of the
    service should have already released it and should not be using it while
    this callback is notified.
    This is True by design for consumers managed by the :class:`Requires`
    decorator in iPOPO, but some others (badly written or very specific design)
    might keep a reference on the service even after its unregistration.

    :Example:

    .. code-block:: python

       @ComponentFactory()
       @Provides("hello.svc")
       class Foo:
           @PostUnregistration
           def callback_method(self, service_reference):
               '''
               service_reference: The ServiceReference of the provided service
               '''
               # Clean up, ...

    :param method: The decorated method
    :raise TypeError: The decorated element is not a valid function
    """
    if not isinstance(method, types.FunctionType):
        raise TypeError("@PostUnregistration can only be applied on functions")

    # Tests the number of parameters
    validate_method_arity(method, "service_reference")
    _append_object_entry(
        method,
        constants.IPOPO_METHOD_CALLBACKS,
        constants.IPOPO_CALLBACK_POST_UNREGISTRATION,
    )
    return method
